// file crcutil.c
#pragma SRC("crcutil.a51")

#include "ecash.h"

void crc_reset()
{
   crcl = 0;
   crch = 0;
}

/*
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: CRC16
;
; Desc:     CRC16 subroutine
;
; Entry:    New byte in ACC, zero crcl/crch to start new computation
;
; Exit:     crcl(bitaddr) and crch(direct) contain crc-16 result
;           ACC is destroyed
*/
void crc16_byte(uchar newByte)
{
   ACC = newByte;
   #pragma asm
      jmp crc16
   #pragma endasm
}

void crc16_block(uchar* block, uchar len)
{
   dbyte cnt = len;
   ACC = block[0];
   do
   {
      #pragma asm
         call crc16
         inc dptr
         movx A, @dptr
      #pragma endasm
   }
   while(--cnt);
}

void crc16(void)
{
   #pragma asm
      xrl A,crcl
      mov crcl,crch
      mov crch,A
      mov C,P
      jnc crc16_1
         xrl crcl,#1
   crc16_1:
      rrc A
      jnc crc16_2
         xrl crcl,#40H
   crc16_2:
      mov c,ACC.7
      xrl A,crch
      rrc A
      mov crch,A
      jnc crc16_3
         xrl crcl,#80H
   crc16_3:
      ret
   #pragma endasm
}


/*
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: CRC8, CRC8First
;
; Desc:     Fast computation of CRC8 w/o iterations of previous routines
;
; Entry:    ACC = New byte to compute into CRC
;
; Exit:     Resulting crc in crcl (DIRECT)
*/
void crc8_byte(uchar newByte)
{
   ACC = newByte;
   #pragma asm
      jmp crc8
   #pragma endasm
}

void crc8(void)
{
   #pragma asm
      push ACC
      xrl A,crcl
      mov crcl,A
      anl crcl,#0F0H
      swap A
      xrl crcl,A
      rr A
      xrl crcl,A
      rr A
      swap A
      anl A,#0C7H
      jnb ACC.2, crc8_1
         cpl ACC.3
   crc8_1:
      xrl crcl,A
      swap A
      anl A,#0CH
      xrl crcl,A
      rr A
      xrl crcl,A
      pop ACC
      ret
   #pragma endasm
}
