//eccmd.c
#include "ecash.h"
#include <string.h>
#include <ctype.h>

#define MAX_RESPONSE_SIZE 128
xbyte ResponseBuffer[MAX_RESPONSE_SIZE];
dbyte ResponseLength;

bit CRCEnabled;
bit CoprocessorNumber;
xbyte PageToReadCopr, PageToReadUser;

void addEndOfLine()
{
   if(CRCEnabled)
   {
      crch = 0;
      crcl = 0;
      crc16_block(ResponseBuffer, ResponseLength);
      byteToHex(~crcl, &ResponseBuffer[ResponseLength]);
      byteToHex(~crch, &ResponseBuffer[ResponseLength+2]);
      ResponseLength += 4;
   }
   ResponseBuffer[ResponseLength] = '\n';
   ResponseLength++;
}

void getDebitTime()
{
   ResponseBuffer[0] = 'R';
   ResponseBuffer[1] = 'D';
   ResponseBuffer[2] = '0';
   ResponseBuffer[3] = '2';
   if(msTimerOverflow)
   {
      ResponseBuffer[4] = 'F';
      ResponseBuffer[5] = 'F';
   }
   else
   {
      byteToHex(msCounter, &ResponseBuffer[4]);
   }
   ResponseLength = 6;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void getFirmwareVersion()
{
   ResponseBuffer[0] = 'R';
   ResponseBuffer[1] = 'V';
   byteToHex(sizeof(FirmwareString), &ResponseBuffer[2]);
   memcpy(&ResponseBuffer[4], FirmwareString, sizeof(FirmwareString));
   ResponseLength = sizeof(FirmwareString) + 3; // (-1 for null character)
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void getCoprocessorRomID()
{
   uchar cnt = 0;
   ResponseBuffer[0] = 'R';
   ResponseBuffer[1] = 'W';
   ResponseBuffer[2] = '1';
   ResponseBuffer[3] = '0';
   if(CoprocessorNumber)
   {
      for(; cnt<8; cnt++)
         byteToHex(COP_RomID_61S[cnt], &ResponseBuffer[2*cnt+4] );
   }
   else
   {
      for(; cnt<8; cnt++)
         byteToHex(COP_RomID[cnt], &ResponseBuffer[2*cnt+4] );
   }
   ResponseLength = 20;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void getStatus()
{
   uchar cnt;
   ResponseBuffer[0] = 'R';
   ResponseBuffer[1] = 'S';
   ResponseBuffer[2] = '5';
   ResponseBuffer[3] = 'E';

   ResponseBuffer[4] = 'A';
   ResponseBuffer[5] = '0';
   ResponseBuffer[6] = '2';
   byteToHex(Mode_Flags, &ResponseBuffer[7]);

   ResponseBuffer[9] = 'C';
   ResponseBuffer[10] = '0';
   ResponseBuffer[11] = '1';
   ResponseBuffer[12] = CRCEnabled?'1':'0';

   ResponseBuffer[13] = 'D';
   ResponseBuffer[14] = '0';
   ResponseBuffer[15] = '2';
   if(msTimerOverflow)
   {
      ResponseBuffer[16] = 'F';
      ResponseBuffer[17] = 'F';
   }
   else
   {
      byteToHex(msCounter, &ResponseBuffer[16]);
   }

   ResponseBuffer[18] = 'F';
   ResponseBuffer[19] = '0';
   ResponseBuffer[20] = '4';
   memcpy(&ResponseBuffer[21], MoneyFileName, 4);

   ResponseBuffer[25] = 'G';
   ResponseBuffer[26] = '0';
   ResponseBuffer[27] = 'F';
   memcpy(&ResponseBuffer[28], LCDMessageDefault, 16);

   ResponseBuffer[44] = 'L';
   ResponseBuffer[45] = '0';
   ResponseBuffer[46] = '2';
   byteToHex(Ack_Flags, &ResponseBuffer[47]);

   ResponseBuffer[49] = 'M';
   ResponseBuffer[50] = '0';
   ResponseBuffer[51] = '6';
   byteToHex(DebitAmount[0], &ResponseBuffer[52]);
   byteToHex(DebitAmount[1], &ResponseBuffer[54]);
   byteToHex(DebitAmount[2], &ResponseBuffer[56]);

   ResponseBuffer[57] = 'N';
   ResponseBuffer[58] = '0';
   ResponseBuffer[59] = '1';
   ResponseBuffer[60] = CoprocessorNumber?'1':'0';;

   ResponseBuffer[61] = 'P';
   ResponseBuffer[62] = '0';
   ResponseBuffer[63] = '2';
   byteToHex(PageToReadCopr, &ResponseBuffer[64]);

   ResponseBuffer[66] = 'T';
   ResponseBuffer[67] = '0';
   ResponseBuffer[68] = '2';
   byteToHex(IOPinToPulse, &ResponseBuffer[69]);

   ResponseBuffer[71] = 'U';
   ResponseBuffer[72] = '0';
   ResponseBuffer[73] = '2';
   byteToHex(PageToReadUser, &ResponseBuffer[74]);

   ResponseBuffer[76] = 'W';
   ResponseBuffer[77] = '1';
   ResponseBuffer[78] = '0';
   if(CoprocessorNumber)
   {
      for(cnt=0; cnt<8; cnt++)
         byteToHex(COP_RomID_61S[cnt], &ResponseBuffer[2*cnt+79] );
   }
   else
   {
      for(cnt=0; cnt<8; cnt++)
         byteToHex(COP_RomID[cnt], &ResponseBuffer[2*cnt+79] );
   }

   ResponseBuffer[95] = 'Z';
   ResponseBuffer[96] = '0';
   ResponseBuffer[97] = '2';
   byteToHex(Status_Flags, &ResponseBuffer[98]);

   ResponseLength = 100;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void getUserPage()
{
   uchar cnt;
   ResponseBuffer[0] = 'R';
   ResponseBuffer[1] = 'U';
   selectCustomer();
   ow_select(); // address the device
   if(SuccessFlag)
   {
      // read memory command
      owBuffer = OW_READ_MEMORY_CMD;
      ow_touchbyte();
      owBuffer = PageToReadUser<<5;
      // send the address
      ow_touchbyte();
      owBuffer = PageToReadUser>>3;
      ow_touchbyte();

      ResponseBuffer[2] = '2';
      ResponseBuffer[3] = '0';
      for(cnt=0; cnt<32; cnt++)
      {
         ow_readbyte();
         byteToHex(owBuffer, &ResponseBuffer[2*cnt+4]);
      }
      ResponseLength = 67;
   }
   else
   {
      ResponseBuffer[2] = '0';
      ResponseBuffer[3] = '0';
      ResponseLength = 4;
   }
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void getCoprPage()
{
   uchar cnt;
   ResponseBuffer[0] = 'R';
   ResponseBuffer[1] = 'P';

   if(CoprocessorNumber)
      memcpy(RomData,COP_RomID_61S,8);
   else
      memcpy(RomData,COP_RomID,8);
   Copr61S = CoprocessorNumber;
   COPSelected = 1;
   NeedAccess = 1;

   ow_select(); // address the device
   if(SuccessFlag)
   {
      // read memory command
      owBuffer = OW_READ_MEMORY_CMD;
      ow_touchbyte();
      owBuffer = PageToReadCopr<<5;
      // send the address
      ow_touchbyte();
      owBuffer = PageToReadCopr>>3;
      ow_touchbyte();

      ResponseBuffer[2] = '2';
      ResponseBuffer[3] = '0';
      for(cnt=0; cnt<32; cnt++)
      {
         ow_readbyte();
         byteToHex(owBuffer, &ResponseBuffer[2*cnt+4]);
      }
      ResponseLength = 68;
   }
   else
   {
      ResponseBuffer[2] = '0';
      ResponseBuffer[3] = '0';
      ResponseLength = 4;
   }
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void setModeFlag()
{
   uchar bitNum = SInBuf[2] - '0';
   if(bitNum<=5)
   {
      uchar enable = 1<<bitNum;
      if(SInBuf[3]=='1')
      {
         // enable mode flag
         Mode_Flags |= enable;
      }
      else
      {
         // disable mode flag;
         Mode_Flags &= ~enable;
      }
   }
}

void setLCDText()
{
   uchar len = hexToByte(&SInBuf[2]);
   memcpy(LCDMessageDefault, &SInBuf[4], 16);
   if(len<16)
      memset(&LCDMessageDefault[len], ' ', 16-len);

   // Only show it if a temp msg isn't currently displayed.
   // otherwise, just wait, it will show up in a second.
   if(!MsgBit)
   {
      lcd_clear();
      lcd_putstring(LCDMessageDefault, 16);
   }

}

void setLockStepAbort()
{
   uchar ackByte = SInBuf[2];
   if((ackByte=='0') && Ack_PollWaiting)
   {
      Ack_PollWaiting = 0;
      Stat_WaitUntilGone = 1;
   }
   else if((ackByte=='1') && Ack_ReadWaiting)
   {
      Ack_ReadWaiting = 0;
      Stat_WaitUntilGone = 1;
   }
   else if((ackByte=='2') && Ack_DebitWaiting)
   {
      Ack_DebitWaiting = 0;
      Stat_WaitUntilGone = 1;
   }
}

void setLockStepAck()
{
   uchar ackByte = SInBuf[2];
   if((ackByte=='0') && Ack_PollWaiting)
   {
      Ack_PollWaiting = 0;
      startTimerSec();
      Stat_Read = 1;
   }
   else if((ackByte=='1') && Ack_ReadWaiting)
   {
      Ack_ReadWaiting = 0;
      startTimerSec();
      Stat_DebitCert = 1;
   }
   else if((ackByte=='2') && Ack_DebitWaiting)
   {
      Ack_DebitWaiting = 0;
      startTimerSec();
      Stat_Poll = 1;
   }
}

void setLockStepEnable()
{
   // Clear all the acknowledgment flags
   Ack_PollWaiting = 0;
   Ack_ReadWaiting = 0;
   Ack_DebitWaiting = 0;

   // Clear ecash state
   Status_Flags = 0;

   switch(SInBuf[2])
   {
      case '0':
         Ack_PollEnabled = (SInBuf[3]=='1')?1:0;
         break;
      case '1':
         Ack_ReadEnabled = (SInBuf[3]=='1')?1:0;
         break;
      case '2':
         Ack_DebitEnabled = (SInBuf[3]=='1')?1:0;
         break;
   }
}

void setPulsePin()
{
   IOPinToPulse = hexToByte(&SInBuf[2]);
   IOPinToPulse = (IOPinToPulse&0x30) | (1<<(IOPinToPulse&0x07));
}

void setIOPinState()
{
   uchar BitNumber = 1<<((SInBuf[3]-'0')&0x07);
   switch(SInBuf[2])
   {
      case '0':
         if(SInBuf[4]=='1')
            P0 |= BitNumber;
         else
            P0 &= ~BitNumber;
         break;
      case '1':
         if(SInBuf[4]=='1')
            P1 |= BitNumber;
         else
            P1 &= ~BitNumber;
         break;
      case '2':
         if(SInBuf[4]=='1')
            P2 |= BitNumber;
         else
            P2 &= ~BitNumber;
         break;
      case '3':
         if(SInBuf[4]=='1')
            P3 |= BitNumber;
         else
            P3 &= ~BitNumber;
         break;
   }
}

void commandProcessor()
{
   if(SerialReady)
   {
      if(SInBuf[0]=='G') // it's a get command
      {
         switch(SInBuf[1])
         {
            // Get last debit time
            case 'D':
               getDebitTime();
               break;

            // Get the last message sent
            case 'L':
               // do nothing will leave message as-is, and re-print
               serial_putstring(ResponseBuffer, ResponseLength);
               break;

            case 'P':
               getCoprPage();
               break;

            case 'S':
               getStatus();
               break;

            case 'U':
               getUserPage();
               break;

            // Get Firmware version
            case 'V':
               getFirmwareVersion();
               break;

            // Get the ROM ID of the selected coprocessor
            case 'W':
               getCoprocessorRomID();
               break;

            // Send back an empty length response, since we don't recognize command
            default:
               ResponseBuffer[0] = 'R';
               ResponseBuffer[1] = SInBuf[1];
               ResponseBuffer[2] = '0';
               ResponseBuffer[3] = '0';
               ResponseLength = 4;
               addEndOfLine();
               serial_putstring(ResponseBuffer, ResponseLength);
               break;
         }
      }
      else if(SInBuf[0]=='S') // it's a set command
      {
         switch(SInBuf[1])
         {
            // enable/disable mode flag
            case 'A':
               setModeFlag();
               break;

            // abort lock-step event
            case 'B':
               setLockStepAbort();
               break;

            // enable/disable the CRC for serial messages
            case 'C':
               CRCEnabled = (SInBuf[2]=='1')?1:0;
               break;

            // set the service filename for user token acct file
            case 'F':
               memcpy(MoneyFileName, &SInBuf[2], 4);
               break;

            // Set the LCD text
            case 'G':
               setLCDText();
               break;

            // set IO Pin to pulse
            case 'I':
               setIOPinState();
               break;

            // acknowledge a lock-step event
            case 'K':
               setLockStepAck();
               break;

            // enable a lock-step event
            case 'L':
               setLockStepEnable();
               break;

            // select a coprocessor for interaction
            case 'N':
               CoprocessorNumber = (SInBuf[2]=='1')?1:0;
               break;

            // Set the page read for coprocessor
            case 'P':
               PageToReadCopr = hexToByte(&SInBuf[2]);
               break;

            // soft reset the ecash processor
            case 'R':
               softReset();
               break;

            // set IO Pin to pulse
            case 'T':
               setPulsePin();
               break;

            // Set the page read for user
            case 'U':
               PageToReadUser = hexToByte(&SInBuf[2]);
               break;

            // play the beeper
            case 'S':
               beep(hexToByte(&SInBuf[2]));
               break;

            // set the serial port reload value
            case 'X':
               TH1 = TL1 = hexToByte(&SInBuf[2]);
               break;
         }

         // Send back an empty length response, to acknowledge command
         ResponseBuffer[0] = 'R';
         ResponseBuffer[1] = SInBuf[1];
         ResponseBuffer[2] = '0';
         ResponseBuffer[3] = '0';
         ResponseLength = 4;
         addEndOfLine();
         serial_putstring(ResponseBuffer, ResponseLength);
      }
      SerialReady = 0;
   }
}

void reportError()
{
   ResponseBuffer[0] = 'E';
   ResponseBuffer[1] = 'E';
   ResponseBuffer[2] = binToHex(errorCode>>4);
   ResponseBuffer[3] = binToHex(errorCode);
   ResponseLength = 4;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void reportToken(bit arrival)
{
   uchar cnt = 0;
   ResponseBuffer[0] = 'E';
   if(arrival)
      ResponseBuffer[1] = 'A';
   else
      ResponseBuffer[1] = 'L';
   ResponseBuffer[2] = '1';
   ResponseBuffer[3] = '0';
   for(; cnt<8; cnt++)
      byteToHex(RomData[cnt], &ResponseBuffer[2*cnt+4] );
   ResponseLength = 20;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void reportRead()
{
   ResponseBuffer[0] = 'E';
   ResponseBuffer[1] = 'S';
   ResponseBuffer[2] = '0';
   ResponseBuffer[3] = '6';
   getAsciiHexBalance(ResponseBuffer, 4);
   ResponseLength = 10;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void reportDebit()
{
   ResponseBuffer[0] = 'E';
   ResponseBuffer[1] = 'D';
   ResponseBuffer[2] = '0';
   ResponseBuffer[3] = '6';
   getAsciiHexBalance(ResponseBuffer, 4);
   ResponseLength = 10;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}

void reportReset()
{
   ResponseBuffer[0] = 'E';
   ResponseBuffer[1] = 'R';
   ResponseBuffer[2] = '0';
   ResponseBuffer[3] = '0';
   ResponseLength = 4;
   addEndOfLine();
   serial_putstring(ResponseBuffer, ResponseLength);
}


