// file ecfuncs.c
#include "ecash.h"

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

; Function:	prepCoprocessor

; Desc:		This is called to read the Co-Processor directory and obtain
;			the page numbers, then compute the addresses that we will use. This
;			must be called at system start-up or when the co-processor is changed.

; Note: For now we will just preset the page addresses and secret addresses. Later
;		we will read the directory and compute these pointers...
*/
void prepCoprocessor(void)
{
   xbyte currentPage = 0;
   xbyte notDoneLoadingFile = 1, notDoneLoadingCopr = 1;
   xbyte found61S = 0, found63S = 0;
   xbyte coprFileName[5] = "COPR ";
   coprFileName[4] = 0x18;

   // init the RomIDs to indicate we haven't found any yet
   *COP_RomID = 0;
   *COP_RomID_61S = 0;
   // Coprocessor's are DS1963S
   DS1963S = 1;
   DS1961S = 0;

   Mode_ODEnabled = 0;
   accessCoprocessor();

   // peform a search
   owResetSearch();
   do
   {
      owNext();
      if(SuccessFlag)
      {
         if(SearchSerialNum[0]!=0x18)
         {
            continue;
         }

         memcpy(RomData, SearchSerialNum, 8);
         // check to see if this is the DS1963S or DS1961S coprocessor
         // by searching for the appropriate file
         currentPage = 0;
         notDoneLoadingFile = 1;
         NeedAccess = 1;
         do
         {
            loadFilePage(currentPage);
            if(SuccessFlag)
            {
               dbyte FilePage = lookupFile(coprFileName);
               if(SuccessFlag)
               {
                  // this coprocessor has the file we are looking for
                  if(coprFileName[4] == 0x33)
                  {
                     found61S = 1;
                     memcpy(COP_RomID_61S, RomData, 8);
                     if(!found63S)
                     {
                        // go ahead and load coprocessor file from first device
                        loadCoprFile(FilePage);
                     }
                     notDoneLoadingFile = 0;
                     // do next search for 63S coprocessor
                     coprFileName[4] = 0x18;
                  }
                  else
                  {
                     found63S = 1;
                     memcpy(COP_RomID, RomData, 8);
                     if(!found61S)
                     {
                        // go ahead and load coprocessor file from first device
                        loadCoprFile(FilePage);
                     }
                     notDoneLoadingFile = 0;
                  }
               }
               else if(FilePage==0)
               {
                  if(coprFileName[4] == 0x33)
                  {
                     // neither file was found on this device
                     // not a valid coprocessor
                     SuccessFlag = 0;
                     // do next search for 63S coprocessor
                     coprFileName[4] = 0x18;
                     notDoneLoadingFile = 0;
                  }
                  else
                  {
                     // start back at directory page
                     currentPage = 0;
                     // look for the other file
                     coprFileName[4] = 0x33;
                  }
               }
               else
               {
                  currentPage = FilePage;
               }
            }
         }
         while(notDoneLoadingFile);


      }
      else
      {
         //no more devices
         notDoneLoadingCopr = 0;
      }
   }
   while(notDoneLoadingCopr);

   // install the binding data to both coprocessors
   if(found61S)
   {
      DS1961S = 1;
      DS1963S = 0;
      NeedAccess = 1;
      accessCoprocessor();
      installBindData();
   }
   if(found63S)
   {
      DS1963S = 1;
      DS1961S = 0;
      NeedAccess = 1;
      accessCoprocessor();
      installBindData();
   }
   Mode_ODEnabled = 1;
}

/*
installs the binding data on the Master Authentication page of the selected token
*/
void installBindData(void)
{
   //Install the Binding Data into the co-processor COP_MAP page -
   eraseScratchpad();
   if(!SuccessFlag)
      return;

   // write the binding data to the scratchpad
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write Scratchpad operation
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_MAP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_MAP>>3;
         ow_touchbyte_crc();
         ow_touchblock_crc(BindData, 32);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Copy SP command
         owBuffer = OW_COPY_SCRATCHPAD_CMD;
         ow_touchbyte();
         owBuffer = COP_MAP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_MAP>>3;
         ow_touchbyte_crc();
         owBuffer = 0x1F; // The E/S Byte
         ow_touchbyte();
         waitUntilDone();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);
}

/*
loads system configuration data from the file that begins on pageNum, off
the currently selected iButton
*/
void loadCoprFile(uchar pageNum)
{
   xbyte index;
   xbyte tempLength, tempIndex, tempCount;

   // load the first page of the file
   // the amount of data will be 29 bytes per page
   NeedAccess = 1;
   loadFilePage(pageNum);
   index = 1;

   // get the name of the money file used on user tokens
   memcpy(MoneyFileName, &ScratchRecord[index], 5);
   index += 5;

   // COP Page for signing certificates (should be 8)
   COP_SDP = ScratchRecord[index];
   index += 1;
   // COP Page for generating Unique Device secrets (usually 7)
   COP_MAP = ScratchRecord[index];
   index += 1;
   // COP Page for storing Unique Device secrets and authenticating (usually 9)
   COP_UDP = ScratchRecord[index];
   COP_UDS = COP_UDP&7; // secret number for page
   index += 1;

   //VersionNumber = ScratchRecord[index];
   index += 1;
   //memcpy(DateCode, ScratchRecord[index], 4);
   index += 4;

   // the 32-byte binding data used to generate unique secrets
   // get the first 15 bytes of the binding data
   memcpy(BindData, &ScratchRecord[index], 15);
   index += 15;

   // load the next page, get from the file continuation pointer
   loadFilePage(ScratchRecord[index]);
   index = 1;

   // get the next 17 bytes of the binding data
   memcpy(&BindData[15], &ScratchRecord[index], 17);
   index += 17;

   // the 7-byte binding code, usually 0xFF for DS1961S compatibility
   memcpy(BindCode, &ScratchRecord[index], 7);
   index += 7;

   // the 3-byte signing challenge, used when signing certificates
   memcpy(SigningChlg, &ScratchRecord[index], 3);
   index += 3;

   // the variable length of the provider name
   ProviderNameLength = ScratchRecord[index];
   index += 1;

   // load the next page, get from the file continuation pointer
   loadFilePage(ScratchRecord[index]);
   index = 1;

   // the variable length of the InitialSig (<20)
   InitialSigLength = ScratchRecord[index];
   index += 1;
   // the variable length of the aux info
   AuxInfoLength = ScratchRecord[index];
   index += 1;

   // Get the name of the service provider
   tempIndex = 0;
   tempLength = ProviderNameLength;
   tempCount = 0;
   do
   {
      if(index==29)
      {
         // load the next page, get from the file continuation pointer
         loadFilePage(ScratchRecord[index]);
         index = 1;
      }
      // figure out how many bytes we can get from this page
      if(tempLength>(29 - index))
         tempCount = 29 - index;
      else
         tempCount = tempLength;
      // get the bytes
      memcpy(&ProviderNameBuffer[tempIndex], &ScratchRecord[index], tempCount);
      // update the indices and the length left to get
      index += tempCount;
      tempIndex += tempCount;
      tempLength -= tempCount;
   }
   while(tempLength>0);

   // initialize the initial signature to all zeros
   memset(InitialSig, 0x00, 20);

   // get the initial signature
   tempIndex = 0;
   tempLength = InitialSigLength;
   tempCount = 0;
   do
   {
      if(index==29)
      {
         // load the next page, get from the file continuation pointer
         loadFilePage(ScratchRecord[index]);
         index = 1;
      }
      // figure out how many bytes we can get from this page
      if(tempLength>(29 - index))
         tempCount = 29 - index;
      else
         tempCount = tempLength;
      // get the bytes
      memcpy(&InitialSig[tempIndex], &ScratchRecord[index], tempCount);
      // update the indices and the length left to get
      index += tempCount;
      tempIndex += tempCount;
      tempLength -= tempCount;
   }
   while(tempLength>0);

   // get the AuxInfo
   tempIndex = 0;
   tempLength = AuxInfoLength;
   tempCount = 0;
   do
   {
      if(index==29)
      {
         // load the next page, get from the file continuation pointer
         loadFilePage(ScratchRecord[index]);
         index = 1;
      }
      // figure out how many bytes we can get from this page
      if(tempLength>(29 - index))
         tempCount = 29 - index;
      else
         tempCount = tempLength;
      // get the bytes
      memcpy(&AuxInfoBuffer[tempIndex], &ScratchRecord[index], tempCount);
      // update the indices and the length left to get
      index += tempCount;
      tempIndex += tempCount;
      tempLength -= tempCount;
   }
   while(tempLength>0);

   SuccessFlag = 1;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	pollForToken
;
; Desc:		See if a token is present. Performs de-bounce to be sure that
; 			token arrival is only reported if sufficient debounce has occurred.
;			This routine is called from MAIN all the time when auto-polling
;			is enabled. Many calls are required without an iButton present
;			before a call will return an arrival state (TC) after a previous
;			arrival has been reported (for debounce).
;
; To use this in a one-shot mode without debounce, set Token_Time to
; zero before calling.
;
; Exit:		Sets SuccessFlag when token of the correct type is present
*/
void pollForToken()
{
   // count polls to gain real entropy
   //Entropy--;
   accessCustomer();
   OWOverdrive = 0;

   // try to read the token romID
   ow_readRomID();
   if(SuccessFlag)
   {
      getDeviceType(); // check the device type
      if(SuccessFlag)
      {
         memcpy(Cust_RomID, RomData, 8);
         SuccessFlag = 1;
         NeedAccess = 0;
         return;
      }
      errorCode = Err_TokenWrongType;
      SuccessFlag = 0;
      return;
   }

   errorCode = Err_NoValidToken;
   // maybe there is a token, and it just doesn't support overdrive
   if(Mode_ODEnabled)
   {
      Mode_ODEnabled = 0;
      ow_readRomID();
      if(SuccessFlag)
      {
         getDeviceType();
         if(!SuccessFlag)
            errorCode = Err_TokenWrongType;
      }
      Mode_ODEnabled = 1;
   }
   SuccessFlag = 0;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: GetDeviceType
;
; Desc:		Examine the family code and determine the device type and if it
;			is valid for us to process.
*/
void getDeviceType(void)
{
   dbyte FamilyCode = RomData[0]&0x7F; // family code minus UniqueWare bit
   DS1963S = 0;
   DS1961S = 0;
   if(FamilyCode==0x18)
   {
      DS1963S = 1;
      SuccessFlag = 1; // set flag for success
   }
   else if(FamilyCode==0x33)
   {
      DS1961S = 1;
      SuccessFlag = 1; // set flag for success
   }
   else
   {
      SuccessFlag = 0;
   }
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	SelectCoProcessor
;
; Desc:		Access the co-processor device
*/
void selectCoprocessor(void)
{
   if(DS1961S)
   {
      DEBUGLN("Select Copr for 61S");
      memcpy(RomData,COP_RomID_61S,8);
      COPSelected = 1;
      if(!Copr61S)
      {
         NeedAccess = 1;
         Copr61S = 1;
      }
   }
   else
   {
      DEBUGLN("Select Copr for 63S");
      memcpy(RomData,COP_RomID,8);
      COPSelected = 1;
      if(Copr61S)
      {
         NeedAccess = 1;
         Copr61S = 0;
      }
   }
}
void accessCoprocessor(void)
{
   if(DS1961S)
   {
      DEBUGLN("Access Copr for 61S");
      memcpy(RomData,COP_RomID_61S,8);
      COPSelected = 1;
      NeedAccess = 1;
      Copr61S = 1;
   }
   else
   {
      DEBUGLN("Access Copr for 63S");
      memcpy(RomData,COP_RomID,8);
      COPSelected = 1;
      NeedAccess = 1;
      Copr61S = 0;
   }
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	SelectCustomer
;
; Desc:		Access the customer device
*/
void selectCustomer(void)
{
   memcpy(RomData,Cust_RomID,8);
   COPSelected = 0;
}
void accessCustomer(void)
{
   memcpy(RomData,Cust_RomID,8);
   COPSelected = 0;
   NeedAccess = 1;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	CheckCRC
;
; Desc:		Read and check the two One Wire inverted CRC16 bytes
;
; Exit:		SuccessFlag = True if CRC is OK
*/
void verifyCRC(void)
{
   ow_readbyte();
   if(crcl == ~owBuffer)
   {
      ow_readbyte();
      if(crch == ~owBuffer)
      {
         crch = 0;
         crcl = 0;
         SuccessFlag = 1;
         return;
      }
   }
   crch = 0;
   crcl = 0;
   SuccessFlag = 0;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: EraseSP
;
; Desc:		Erase the scratch pad only in the current device
;
; Exit:		TA1, TA2, and E/S unpredictable
*/
void eraseScratchpad(void)
{
   DEBUGLN("      Erasing scratcphad");
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         owBuffer = OW_ERASE_SCRATCHPAD_CMD; // erase scratchpad
         ow_touchbyte();
         owBuffer = 0x00;
         ow_touchbyte();
         owBuffer = 0x00;
         ow_touchbyte();
         waitUntilDone();
         if(SuccessFlag)
         {
            DEBUGLN("      Scratchpad erased");
            return;
         }
      }
      NeedAccess = 1;
      SuccessFlag = 0;
   }
   while(--Retry);
   DEBUGLN("      Scratchpad erase failed");
   SuccessFlag = 0;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	WaitUntilDone
;
; Desc:		Loop getting bits until we read a zero or 256 bits go by
;
; Exit:		TC if all is well, FC if time-out
;
*/
void waitUntilDone(void)
{
   // wait some time first ?????????????????????????????????
   //waitCounter = 0;
   //while(--waitCounter);

   // wait for bit
   dbyte waitCounter = 0;
   do
   {
      owBufferOut = 1;
      ow_touchbit();
      if(!owBufferIn)
      {
         SuccessFlag = 1;
         return;
      }
   }
   while(--waitCounter);
   SuccessFlag = 0;
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	readToken
;
; Desc:		Read the present token and validate the balance in it
;
; Exit:		ACC=0 if AOK, else A=Error Code
*/
void readToken(void)
{
   errorCode = 0;
   DEBUGLN("Read Token Called");
   formatCertificate(); // init the certificate prototype

   selectCustomer();
   openAccount(); // locate the correct device page
   if(!SuccessFlag)
   {
      errorCode = Err_OpenAccount;
      DEBUGLN("openAccount failed");
      return;
   }

   // use access to make sure we have the right coprocessor selected
   accessCoprocessor(); // select the coprocessor device

   getChallenge(); // get a challenge from the coprocessor device
   if(!SuccessFlag)
   {
      errorCode = Err_GetChallenge;
      DEBUGLN("getChallenge failed");
      return;
   }

   selectCustomer();
   issueChallenge();
   if(!SuccessFlag)
   {
      errorCode = Err_IssueChallengeR;
      DEBUGLN("issueChallenge failed");
      return;
   }


   readAccountPage();
   if(!SuccessFlag)
   {
      DEBUGLN("readAccountPage failed");
      errorCode = Err_ReadAccount;
      return;
   }

   verifyFile();
   if(!SuccessFlag)
   {
   }

   memcpy(OldRecord, Certificate.Record.ptr, 32+4);

   selectCoprocessor();
   computeDeviceSecret(); // figure out the device unique secret
   if(!SuccessFlag)
   {
      errorCode = Err_CompDevSecret;
      DEBUGLN("computeDeviceSecret failed");
      return;
   }

   authenticate();
   if(!SuccessFlag)
   {
      if(DS1961S)
      {
         selectCustomer();
         fixer();
      }
      errorCode = Err_Authenticate;
      DEBUGLN("authenticate failed");
      return;
   }


   if(DS1963S)
   {
      validate();
      if(!SuccessFlag)
      {
         errorCode = Err_Validate;
         DEBUGLN("validate failed");
         return;
      }
   }
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	FormatCertificate
;
; Desc:		Prepare the certifcate prototype for use given the device Rom ID
*/
void formatCertificate(void)
{
   memset(Certificate.Record.ptr, 32, 0); // zero the entire certificate prototype
   memcpy(Certificate.DeviceROMID, Cust_RomID, 8); // Install the ROM ID
   memcpy(Certificate.SerialNumber, Cust_RomID, 7); // Install the Serial Number
   memcpy(Certificate.FileName, MoneyFileName, 5); // set the file name
   Certificate.Descriptor[0] = 1;
   Certificate.Descriptor[1] = 5; // install data object descriptor
   Certificate.ObjectLength = sizeof(Certificate); // install data object length
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	OpenAccount
;
; Desc:		Open the token & account specified in the current certificate and
;			prepare to access the account file.
*/
void openAccount(void)
{
   dbyte filePage;
   DEBUGLN("   Open Account Called");
   Certificate.DevicePage = 0;
   for(;;)
   {
      loadFilePage(Certificate.DevicePage);
      if(SuccessFlag)
      {
         DEBUGLN("      loadFilePage success");
         filePage = lookupFile(Certificate.FileName);
         if(SuccessFlag)
         {
            DEBUGLN("      lookupFile success");
            Certificate.DevicePage = filePage;
            filePage = filePage & 0x3F;
            // remember page number in the eCert
            Certificate.PageNumber = filePage;
            // remember pge number in direct memory
            CUST_ECP = filePage;
            return;
         }
         else if(!filePage)// check if no more directory
         {
            return;
         }
         else
         {
            // go ahead and load the next page of the directory
            Certificate.DevicePage = filePage;
         }
      }
      else
      {
         // no directory
         return;
      }
   }
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: LoadFilePage

; Desc:		Given a customer device page number at DevicePage, read
;			that page into ScratchRecord. Retry to get valid CRC16. Return
;			SuccessFlag=true if successful load.
;
*/
void loadFilePage(uchar pageNumber)
{
   dbyte fileLength;
   //specificErrorCode = Err_Spec_None;

   Retry = RetryCount;
   do
   {
      ow_select(); // address the device
      if(SuccessFlag)
      {
         // read memory command
         owBuffer = OW_READ_AUTH_PAGE_CMD;
         ow_touchbyte_crc();
         // send the address
         owBuffer = pageNumber<<5;
         ow_touchbyte_crc();
         owBuffer = pageNumber>>3;
         ow_touchbyte_crc();

         ow_readblock_crc(ScratchRecord, 32);
         if(DS1963S)
         {
            sendNulls(8);
         }
         else if(DS1961S)
         {
            sendNulls(1);
         }
         verifyCRC();
         msDelay(2);
         sendNulls(1);
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
   {
      //specificErrorCode = Err_Spec_Bad_CRC_On_Cmd;
      return;
   }

   // save the File length
   fileLength = ScratchRecord[0]+1;
   // seed the crc16 with the pagenumber
   crcl = pageNumber;
   crch = 0;
   crc16_block(ScratchRecord, fileLength);

   // get the File CRC
   if((~crcl)==ScratchRecord[fileLength])
   {
      if((~crch)==ScratchRecord[fileLength+1])
      {
         SuccessFlag = 1;
         return;
      }
   }
   //specificErrorCode = Err_Spec_Bad_Directory;
   SuccessFlag = 0;
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: LookUpFile
;
; Desc:		Seek a file name match in the loaded directory record at ScratchRecord.
;
; Exit:		TC if file name found and ACC = Starting page number
;			FC if failed to find name and ACC = CP (next directory page)
;			If FC and ACC=0, then file not found in device at all.
;
*/
uchar lookupFile(uchar *fileName)
{
   dbyte index = 1;

   // Make sure it's a directory length
   if(ScratchRecord[0]>29)
   {
      SuccessFlag = 0;
      return 0;
   }

   for(;;)
   {
      // chec if this is the entry we are looking for
      if(memcmp(&ScratchRecord[index], fileName, 5)==0)
      {
         SuccessFlag = 1;
         return ScratchRecord[index+5];
      }
      // advance to the next entry
      index += 7;
      if(index==29)
      {
         // get the next directory page
         SuccessFlag = 0;
         return ScratchRecord[ScratchRecord[0]];
      }
   }
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	GetChallenge
;
; Desc:		Uses the token specified by the current certificate to compute a
;			random challenge value in that certificate. Always uses Secret #1
;			in whatever device is specified (usually the co-processor).
*/
void getChallenge()
{
   // ensure access to the scratchpad
   eraseScratchpad();
   if(!SuccessFlag)
      return;

   Retry = RetryCount;
   do
   {
      // have the coprocessor generate a random challenge
      ow_select();
      if(SuccessFlag)
      {
         // SHA Function command
         owBuffer = OW_COMPUTE_SHA_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_MAP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_MAP>>3;
         ow_touchbyte_crc();
         // compute challenge command
         owBuffer = OW_COMPUTE_CHALLENGE_CMD;
         ow_touchbyte_crc();
         verifyCRC();
         if(SuccessFlag)
         {
            waitUntilDone();
            if(SuccessFlag)
            {
               break;
            }
         }
      }
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   Retry = RetryCount;
   do
   {
      // Read the challenge bytes and store at Challenge
      ow_select();
      if(SuccessFlag)
      {
         // Read Scratchpad command
         owBuffer = OW_READ_SCRATCHPAD_CMD;
         ow_touchbyte();
         sendNulls(11); // Read the address and E/S bytes, and skip to MAC
         ow_readblock(Certificate.Challenge, 3);
         // DEBUG: Let's use constant challenge
         //Certificate.Challenge[0] = 0xCD;
         //Certificate.Challenge[1] = 0xD4;
         //Certificate.Challenge[2] = 0x50;
         DEBUGBLOCKLBL("   getChallenge: ", Certificate.Challenge, 3);
         return;
      }
      NeedAccess = 1;
   }
   while(--Retry);

   //TODO: Possibly add in real-world entropy

}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	IssueChallenge
;
; Desc:		Sends the challenge value provided to the token device in preparation
;			for an authenticated data read. Assumes that the device may be a new
;			arrival and clears SP first.
;
; Entry:	DeviceRomID
;			DeviceChannel
;			Challenge (copied from co-processor certificate)
;
; Exit:		Error = 0 if all is well
;			Challenge
;			SuccessFlag = Set if all is well
*/
void issueChallenge()
{
   if(DS1963S)
   {
      // erase scratchpad for DS1963S
      eraseScratchpad();
      if(!SuccessFlag)
      {
         DEBUGLN("issueChallenge: eraseScratchpad failed");
         return;
      }
   }

   // Send the challenge bytes to the customer device -
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write Scratchpad command
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();

         // send the address
         if(DS1963S)
         {
            owBuffer = (CUST_ECP<<5)+20; // start of challenge in DS1963S
            ow_touchbyte_crc();
            owBuffer = CUST_ECP>>3;
            ow_touchbyte_crc();
         }
         else if(DS1961S)
         {
            owBuffer = CUST_ECP<<5;
            ow_touchbyte_crc();
            owBuffer = CUST_ECP>>3;
            ow_touchbyte_crc();
            sendNulls(4); // Start of challenge in DS1961S
         }

         //write out the challenge
         ow_touchblock_crc(Certificate.Challenge, 3);

         if(DS1963S)
         {
            // fill out the DS1963S scratchpad
            sendNulls(9);
         }
         else if (DS1961S)
         {
            // fill out the DS1961S scratchpad
            owBuffer = 0xFF;
            ow_touchbyte_crc();
         }

         verifyCRC();
         if(SuccessFlag)
         {
            DEBUGLN("issueChallenge: CRC Check successful");
            return;
         }
      }
      NeedAccess = 1;
   }
   while(--Retry);

   SuccessFlag = 0;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	ReadAccount
;
; Desc:		Reads the account data into the certificate from the device
;			and file page specified in the certificate and reads the MAC
;			into the certificate as well. Checks the length and MUC
;			values and the certificate File CRC-16.
*/
void readAccountPage(void)
{
   uchar FileLength = 0;

   Retry = RetryCount;
   do
   {
      // Issue the Read Auth Page command and read in the page data
      ow_select();
      if(SuccessFlag)
      {
         // Read Auth Page Command
         owBuffer = OW_READ_AUTH_PAGE_CMD;
         ow_touchbyte_crc();
         owBuffer = CUST_ECP<<5;
         ow_touchbyte_crc();
         owBuffer = CUST_ECP>>3;
         ow_touchbyte_crc();

         // Read the certificate
         if(DS1963S)
         {
            // 32-byte certificate + 4-byte write-cycle counter
            ow_readblock_crc(Certificate.Record.ptr, 32+4);
            // 4-byte secret write-cycle counter
            ow_readblock_crc((uchar*)&Certificate.SecretCounter, 4);

            //PRINTBLOCKLBL("readAccountPage: file:", Certificate.Record.ptr, 32);
            //PRINTBLOCKLBL("readAccountPage: WCC:", Certificate.WriteCounter, 4);
            //PRINTBLOCKLBL("readAccountPage: SCC:", Certificate.SecretCounter, 4);
         }
         else if(DS1961S)
         {
            // clear the cycle counters
            memset(Certificate.WriteCounter, 0xFF, 4);
            memset(Certificate.SecretCounter, 0xFF, 4);
            // 32-byte certificate
            ow_readblock_crc(Certificate.Record.ptr, 32);
            owBuffer = 0xFF; // read the dummy byte
            ow_touchbyte_crc();
            //PRINTBLOCKLBL("readAccountPage: file:", Certificate.Record.ptr, 32);
         }
         verifyCRC();
         if(SuccessFlag)
         {
            if(DS1963S)
               waitUntilDone();
            else if(DS1961S)
               wait2ms();
         }
         if(SuccessFlag)
            break;
      }
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   //PRINTLN("readAccountPage: Reading MAC");
   Retry = RetryCount;
   do
   {
      crc_reset();
      // read MAC
      if(DS1961S)
      {
         ow_readblock_crc(Certificate.AuthMAC, 20);
         //PRINTBLOCKLBL("readAccountPage: AuthMAC:", Certificate.AuthMAC, 20);
      }
      else if(DS1963S)
      {
         ow_select();
         if(SuccessFlag)
         {
            // do a dummy write SP to set TA1/TA2
            owBuffer = OW_WRITE_SCRATCHPAD_CMD;
            ow_touchbyte();
            owBuffer = 0x08; // point to first byte of MAC in spad
            ow_touchbyte();
            owBuffer = 0x00;
            ow_touchbyte();
            // re-select the device
            ow_select();
            if(SuccessFlag)
            {
               // do a read scratchpad for MAC
               owBuffer = OW_READ_SCRATCHPAD_CMD;
               ow_touchbyte_crc();
               sendNulls(3); // Read TA1/TA2, E/S
               // read the 20-byte MAC from the spad
               ow_readblock_crc(Certificate.AuthMAC, 20);
               //PRINTBLOCKLBL("readAccountPage: AuthMAC:", Certificate.AuthMAC, 20);
               sendNulls(4); // discard four more bytes
            }
         }
      }
      verifyCRC();
      if(SuccessFlag)
      {
         //PRINTLN("readAccountPage: MAC CRC success");
         break;
      }
      else
      {
         //PRINTLN("readAccountPage: MAC CRC failed");
      }
      NeedAccess = 1;
   }
   while(--Retry);
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	ComputeDeviceSecret
;
; Desc:		Given a device serial number at DeviceRomID, compute the unique
;			device secret in the co-processor (secret 2).
*/
void computeDeviceSecret(void)
{
   dbyte secretOffset = COP_UDS<<3;
   dbyte secretEndOffset = (secretOffset+7)&0x1F;

   eraseScratchpad();
   if(!SuccessFlag)
      return;

   //Write the Bind Code, Page Number, and ROM ID to the SP -
   DEBUGLN("Computing Device Secret");
   /*DEBUGLN("   Write Scratchpad:");
   DEBUGBLOCKLBL("      Page#=", &COP_MAP, 1);
   temp1 = (COP_MAP<<5) + 8;
   DEBUGBLOCKLBL("      TA2=", &temp1, 1);
   temp1 = COP_MAP>>3;
   DEBUGBLOCKLBL("      TA1=", &temp1, 1);
   DEBUGLN("      Data: ");
   DEBUGBLOCK(BindCode, 4);
   DEBUGBLOCK(&Certificate.DevicePage, 1);
   DEBUGBLOCK(Certificate.SerialNumber, 7);
   DEBUGBLOCK(BindCode+4, 3);
   DEBUGLN("");*/

   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write scratchpad operation
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = (COP_MAP<<5) + 8; // offset = 8
         ow_touchbyte_crc();
         owBuffer = COP_MAP>>3;
         ow_touchbyte_crc();
         // Write the Bind Code[0-3]
         ow_touchblock_crc(BindCode, 4);
         // Page Number
         owBuffer = CUST_ECP;
         ow_touchbyte_crc();
         // ROM ID
         ow_touchblock_crc(Certificate.SerialNumber, 7);
         // and Bind Code[4-7]
         ow_touchblock_crc(BindCode+4, 3);
         sendNulls(9); // fill to end for CRC
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("   Write scratchpad success");

   // compute next secret
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // SHA Function command
         owBuffer = OW_COMPUTE_SHA_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_MAP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_MAP>>3;
         ow_touchbyte_crc();
         // Compute next secret command
         owBuffer = OW_COMPUTE_NEXT_SECRET_CMD;
         ow_touchbyte_crc();
         verifyCRC();
         if(SuccessFlag)
            waitUntilDone();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("Compute Next Secret success");

   // copy scratchpad to the appropriate secret
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write Scratchpad, to set address
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte();
         owBuffer = secretOffset;
         ow_touchbyte();
         owBuffer = 2; // all secrets start at 0x200;
         ow_touchbyte();
      }
      ow_select();
      if(SuccessFlag)
      {
         // Copy Scratchpad command
         owBuffer = OW_COPY_SCRATCHPAD_CMD;
         ow_touchbyte();
         owBuffer = secretOffset;
         ow_touchbyte();
         owBuffer = 2; // all secrets start at 0x200;
         ow_touchbyte();
         owBuffer = secretEndOffset;
         ow_touchbyte();
         waitUntilDone();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("Copy scratchpad success");

   // for DS1961S, need to write secret zero as well
   if(DS1961S)
   {
      DEBUGLN("Copying secret to secret 0");

      // copy scratchpad to the secret 0
      Retry = RetryCount;
      do
      {
         ow_select();
         if(SuccessFlag)
         {
            // Write Scratchpad, to set address
            owBuffer = OW_WRITE_SCRATCHPAD_CMD;
            ow_touchbyte();
            owBuffer = 0; // secret 0
            ow_touchbyte();
            owBuffer = 2; // all secrets start at 0x200;
            ow_touchbyte();
         }
         ow_select();
         if(SuccessFlag)
         {
            // Copy Scratchpad command
            owBuffer = OW_COPY_SCRATCHPAD_CMD;
            ow_touchbyte();
            owBuffer = 0; // secret 0
            ow_touchbyte();
            owBuffer = 2; // all secrets start at 0x200;
            ow_touchbyte();
            owBuffer = 0x07; // The E/S byte
            ow_touchbyte();
            waitUntilDone();
         }
         if(SuccessFlag)
            return;
         NeedAccess = 1;
      }
      while(--Retry);
   }
}



/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	Authenticate
;
; Desc:		Given a certificate and MAC recently read from a token, use the
;			co-processor to authenticate the token and data from it.
*/
void authenticate(void)
{
   if(DS1961S)
   {
      // Null W/C Counter
      memset(Certificate.WriteCounter,0xFF,4);
   }

   // assure access to the scratchpad
   eraseScratchpad();
   if(!SuccessFlag)
      return;

   // write the file to the co-processor
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write scratchpad operation
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_UDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_UDP>>3;
         ow_touchbyte_crc();
         ow_touchblock_crc(Certificate.Record.ptr, 32);
         verifyCRC();
         if(SuccessFlag)
            break;
      }
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("   write scratchpad success");

   // issue copy scratchpad
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Copy scratchpad operation
         owBuffer = OW_COPY_SCRATCHPAD_CMD;
         ow_touchbyte();
         owBuffer = COP_UDP<<5;
         ow_touchbyte();
         owBuffer = COP_UDP>>3;
         ow_touchbyte();
         owBuffer = 0x1F; // E/S byte
         ow_touchbyte();
         waitUntilDone();
         if(SuccessFlag)
            break;
      }
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("File copy to Coprocessor success");

   // send SP data to copr scratchpad
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         owBuffer = OW_WRITE_SCRATCHPAD_CMD; // Write scratchpad operation
         ow_touchbyte_crc();
         owBuffer = (COP_UDP<<5) + 8; // start at the 8th byte
         ow_touchbyte_crc();
         owBuffer = COP_UDP>>3;
         ow_touchbyte_crc();
         // write counter, S/N, page#, and challenge
         ow_touchblock_crc(Certificate.WriteCounter, 15);
         sendNulls(9); // fill out the page
         verifyCRC();
         if(SuccessFlag)
            break;
      }
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("Scratchpad signing data write success");

   // Sign the data
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // SHA Function command
         owBuffer = OW_COMPUTE_SHA_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_UDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_UDP>>3;
         ow_touchbyte_crc();
         if(DS1963S)
            // Validate Data Page command
            owBuffer = OW_VALIDATE_DATA_PAGE_CMD;
         else if(DS1961S)
            // Authenticate Host command
            owBuffer = OW_AUTHENTICATE_HOST_CMD;
         ow_touchbyte_crc();
         verifyCRC();
         if(SuccessFlag)
            waitUntilDone();
         if(SuccessFlag)
            break;
      }
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("SHA Function success");

   // Match the scratchpad
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Match Scratchpad command
         owBuffer = OW_MATCH_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         ow_touchblock_crc(Certificate.AuthMAC, 20);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("Match scratchpad CRC success");
   waitUntilDone();
   if(SuccessFlag)
      return;

   owBufferOut = 1;
   ow_touchbit();
   if(owBufferIn)
      SuccessFlag = 0;
   else
      SuccessFlag = 1;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	Validate
;
; Desc:		Validate the monetary balance in the certificate
;
; Entry:
;
; Exit:		Error = 0 if all is well
*/
void validate(void)
{
   DEBUGLN("   Validating money file");

   memcpy(ScratchRecord, Certificate.Record.rec.MAC, 20);

   setupForSignature();
   if(!SuccessFlag)
      return;

   // send the validata data page command
   DEBUGLN("   Validate data page");
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // start a SHA command
         owBuffer = OW_COMPUTE_SHA_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_SDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         // Perform Validate Data Page
         owBuffer = OW_VALIDATE_DATA_PAGE_CMD;
         ow_touchbyte_crc();
         verifyCRC();
         if(SuccessFlag)
            waitUntilDone();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
     return;

   DEBUGLN("   Validata data page success");

   // Perform SP Match command w/MAC (Monetary MAC)
   //PRINTLN("   Matching Money MAC");
   //PRINTBLOCKLBL("      Money MAC: ", ScratchRecord, 20);
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Do Match Scratchpad
         owBuffer = OW_MATCH_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         ow_touchblock_crc(ScratchRecord, 20);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(SuccessFlag)
   {
      waitUntilDone();
      if(SuccessFlag)
      {
         DEBUGLN("   Match Scratchpad succeeded");
      }
      else
      {
         DEBUGLN("   Match Scratchpad failed");
      }
   }
   else
   {
      DEBUGLN("   Match Scratchpad CRC failed");
   }
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	SetupSignature
;
; Desc:		Prepare the co-processor to validate or sign the certificate
;			(Used by Validate and by SignCertificate).
;
; Notes:	When we sign an eCertificate, we must replace the old signature
;			and the CRC16 bytes with zeros, and then compute the signature.
*/
void setupForSignature(void)
{
   DEBUGLN("setupForSignature Called");
   // insert the initial signing data
   memcpy(Certificate.Record.rec.MAC, InitialSig, 20);
   // zero out the file CRC
   Certificate.Record.rec.FileCRC_l = 0;
   Certificate.Record.rec.FileCRC_h = 0;

   // ensure access to the scratchpad
   eraseScratchpad();
   if(!SuccessFlag)
      return;

   DEBUGLN("   Writing certificate file to signing page");
   DEBUGBLOCKLBL("      Certificate to be signed looks like: ", Certificate.Record.ptr, 32);

   // Write this record to the SP for copy to memory page -
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // write scratchpad
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_SDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         ow_touchblock_crc(Certificate.Record.ptr, 32);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("   Write scratchpad succeeded.");

   // Copy the SP data to the page for signing -
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // copy scratchpad
         owBuffer = OW_COPY_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_SDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         owBuffer = 0x1F;
         ow_touchbyte_crc();
         waitUntilDone();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   DEBUGLN("   Copy scratchpad succeeded.");

   // Send SP data to co-processor scratchpad -
   DEBUGLN("   Write scratchpad with signing data");
   DEBUGBLOCKLBL("      counter, page, RomID: ", Certificate.WriteCounter, 12);
   DEBUGBLOCKLBL("      SigningChlg: ", SigningChlg, 3);
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // write scratchpad
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = (COP_SDP<<5) + 8;  // start at 8th byte
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         // send counter(4), page#(1), and RomID(7)
         ow_touchblock_crc(Certificate.WriteCounter, 12);
         // send the signing challenge
         ow_touchblock_crc(SigningChlg, 3);
         // fill out to get crc
         sendNulls(9);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(SuccessFlag)
   {
      DEBUGLN("   Write scratchpad with signing data succeeded.");
   }
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	Debit_Cert, Revalue_Cert
;
; Desc:		Perform a debit of the eCert balance field by the amount in
;			DebitAmount. revalueToken adds the current DebitAmount
;			value to Balance. Computes a new transaction ID, recomputes
;			the eCert CRC, and re-signs the eCert to make it ready to write
;			to the customer token.
;
; Enter:	DebitAmount (direct) contains debit or revalue amount in
;			little-endian hexadecimal minor units (pennys).
;
; Exit:		ACC=0 if AOK, else A=Error Code
;			This should be followed by a readBackVerify if successful.
`*/
void debitCertificate(void)
{
   if(DS1963S)
   {
      uchar tempBalance[3];
      debitBalance(Certificate.Record.rec.Balance, tempBalance);
      if(SuccessFlag)
         memcpy(Certificate.Record.rec.Balance, tempBalance, 3);
   }
   else if(DS1961S)
   {
      DEBUGLN("Debiting DS1961S Certificate");
      if(Certificate.Record.rec33.FileLength==RECORD_A_LENGTH)
      {
         DEBUGLN("   Was Record A, switch to Record B");
         debitBalance(Certificate.Record.rec33.Balance_A, Certificate.Record.rec33.Balance_B);
         if(SuccessFlag)
            Certificate.Record.rec33.FileLength = RECORD_B_LENGTH;
      }
      else
      {
         DEBUGLN("   Was Record B, switch to Record A");
         debitBalance(Certificate.Record.rec33.Balance_B, Certificate.Record.rec33.Balance_A);
         if(SuccessFlag)
            Certificate.Record.rec33.FileLength = RECORD_A_LENGTH;
      }
   }

   if(SuccessFlag)
   {
      updateCertificate();
   }
   else
   {
      DEBUGLN("Debit Certificate failed");
   }
}

void updateCertificate(void)
{
   uchar wcc[4]; // save the write-cycle counter value
   memcpy(wcc, Certificate.WriteCounter, 4);
   // generate a new transaction ID
   if(DS1963S)
   {
      makeNewTransID(Certificate.Record.rec.TransID);
      incPageCounter();
      signCertificate();
   }
   else if(DS1961S)
   {
      if(Certificate.Record.rec33.FileLength==RECORD_A_LENGTH)
      {
         makeNewTransID(Certificate.Record.rec33.TransID_A);
      }
      else
      {
         makeNewTransID(Certificate.Record.rec33.TransID_B);
      }
   }
   // ensure continuation pointer is correct
   Certificate.Record.ptr[Certificate.Record.rec.FileLength] = 0;
   // fix-up the File CRC
   computeFileCRC();

   // save a copy for future reference
   memcpy(NewRecord, Certificate.Record.ptr, 36);
   // restore original value, in case we need to retry
   memcpy(Certificate.WriteCounter, wcc, 4);
   SuccessFlag = 1;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	SignCertificate
;
; Desc:		Given an altered certificate, generate the proper monetary
;			MAC in it for the new values. This routine sends zeros for the
;			MAC and the File CRC bytes. This expects the page write counter
;			to be incremented.
;
; Note:		This increments the 32-bit WriteCounter store first.
;
; Exit:		Error = 0 if all is well.
; 			MAC is filled in
;
*/
void signCertificate(void)
{
   DEBUGLN("signCertificate Called");
   selectCoprocessor();
   setupForSignature();
   if(!SuccessFlag)
      return;

   DEBUGLN("   Signing the certificate");

   // do Sign Data Page
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // start SHA function
         owBuffer = OW_COMPUTE_SHA_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_SDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         // Perform Sign Data Page
         owBuffer = OW_SIGN_DATA_PAGE_CMD;
         ow_touchbyte_crc();
         verifyCRC();
         if(SuccessFlag)
            waitUntilDone();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   // get the signature and store in MAC
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // do a dummy write SP
         // write scratchpad command
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte();
         owBuffer = 0x08; // offset 8, for MAC
         ow_touchbyte();
         owBuffer = 0x00;
         ow_touchbyte();

         // read the scratchpad
         ow_select();
         if(SuccessFlag)
         {
            // read scratchpad command
            owBuffer = OW_READ_SCRATCHPAD_CMD;
            ow_touchbyte_crc();
            sendNulls(3); // discard TA1/TA2, E/S
            ow_readblock_crc(Certificate.Record.rec.MAC, 20);
            sendNulls(4); // to get to the end
            verifyCRC();
         }
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(SuccessFlag)
   {
      DEBUGLN("   Certificate signed successfully");
   }
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	writeAccountPage
;
; Desc:		Write the debited account to the iButton.
;           Note that success is assumed once the Copy Scratchpad command
;           is performed and no retries will be attempted after that point.
;
; Exit:		This should be followed by a readBackVerify if successful.
*/
void writeAccountPage(void)
{
   DEBUGLN("   Writing Debited Account");
   selectCustomer();
   if(DS1963S)
   {
      // ensure access to scratchpad
      eraseScratchpad();
      if(!SuccessFlag)
      {
         errorCode = Err_EraseScratchpad;
         return;
      }

      Retry = RetryCount;
      do
      {
         ow_select();
         if(SuccessFlag)
         {
            // Write scratchpad command
            owBuffer = OW_WRITE_SCRATCHPAD_CMD;
            ow_touchbyte_crc();
            owBuffer = CUST_ECP<<5;
            ow_touchbyte_crc();
            owBuffer = CUST_ECP>>3;
            ow_touchbyte_crc();
            // write the entire eCertificate
            ow_touchblock_crc(Certificate.Record.ptr, 32);
         }
         if(SuccessFlag)
            break;
         NeedAccess = 1;
      }
      while(--Retry);

      if(SuccessFlag)
      {
         Retry = RetryCount;
         do
         {
            readScratchpadBack(0); // check it byte for byte
            if(SuccessFlag)
               break;
            NeedAccess = 1;
         }
         while(--Retry);

         if(SuccessFlag)
         {
            // can't retry copySP operation, because of page counter
            //doCopyScratchpad63S(CUST_ECP, 0);
            ow_select();
            if(SuccessFlag)
            {
               owBuffer = OW_COPY_SCRATCHPAD_CMD; // copy scratchpad
               ow_touchbyte();
               owBuffer = CUST_ECP<<5;
               ow_touchbyte();
               owBuffer = CUST_ECP>>3;
               ow_touchbyte();
               owBuffer = 0x1F; // E/S byte
               ow_touchbyte();
               waitUntilDone();
            }
         }
      }

      if(SuccessFlag)
      {
         // increment page counter to represent new value
         incPageCounter();
      }

   }
   else if(DS1961S)
   {
      memcpy(ScratchRecord, OldRecord, 36);
      if(Certificate.Record.rec33.FileLength==RECORD_B_LENGTH)
      {
         ////PRINTLN("Updating Record B");
         // update record B
         updateAccountPage61S(16);
         if(SuccessFlag)
         {
            ////PRINTLN("Updating Header");
            memcpy(&ScratchRecord[16], &Certificate.Record.ptr[16], 8);
            // update the header
            updateAccountPage61S(0);
         }
      }
      else
      {
         ////PRINTLN("Updating Record A");
         // update Record A
         updateAccountPage61S(8);
         if(SuccessFlag)
         {
            ////PRINTLN("Updating Header");
            memcpy(&ScratchRecord[8], &Certificate.Record.ptr[8], 8);
            // update the header
            updateAccountPage61S(0);
         }
      }

      if(!SuccessFlag)
      {
         // TODO: Refresh part here
         fixer();
      }
   }

   if(!SuccessFlag)
   {
      errorCode = Err_WriteAccountPage;
   }
}

void updateAccountPage61S(uchar offset)
{
   ////PRINTBLOCKLBL("      Updating Account Page, offset=", &offset, 1);
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write Scratchpad
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = (CUST_ECP<<5) + offset; // address the proper 8 bytes
         ow_touchbyte_crc();
         owBuffer = CUST_ECP>>3;
         ow_touchbyte_crc();
         // write the balance, TID, CP, and CRC-16
         ow_touchblock_crc(&Certificate.Record.ptr[offset], 8);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      //PRINTLN("Accessing");
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   //PRINTLN("Checking scratchpad");

   Retry = RetryCount;
   do
   {
      // check it byte for byte
      readScratchpadBack(offset);
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   //PRINTLN("Get copy SP MAC");

   // get CopySP MAC for updating
   selectCoprocessor();
   getCopyScratchpadMAC(offset);
   if(!SuccessFlag)
      return;

   //PRINTLN("do copy SP MAC");

   // do CopySpad
   selectCustomer();
   doCopyScratchpad61S(CUST_ECP, offset);
   //if(!SuccessFlag)
   //{
      //PRINTLN("   Copy Scratchpad 61S Failed");
   //}
}

// check the scratchpad contents, byte for byte
void readScratchpadBack(uchar offset)
{
   dbyte temp;
   dbyte len;
   uchar *compareData;

   DEBUGLN("      Reading/Comparing Scratchpad contents");

   ow_select();
   if(!SuccessFlag)
      return;

   // Read Scratchpad
   owBuffer = OW_READ_SCRATCHPAD_CMD;
   ow_touchbyte_crc();

   // check TA2
   if(DS1963S)
   {
      temp = CUST_ECP<<5;
   }
   else if(DS1961S)
   {
      temp = (CUST_ECP<<5) + offset; // address the last 8 bytes
   }
   ow_readbyte();
   if(owBuffer!=temp)
   {
      DEBUGLN("         Bad TA2");
      SuccessFlag = 0;
      return;
   }

   // check TA1
   temp = CUST_ECP>>3;
   ow_readbyte();
   if(owBuffer!=temp)
   {
      DEBUGLN("         Bad TA1");
      SuccessFlag = 0;
      return;
   }

   // check E/S
   ow_readbyte();
   if(DS1963S)
   {
      if(owBuffer!=31)
      {
         DEBUGLN("         Bad E/S");
         SuccessFlag = 0;
         return;
      }
      len = 32;
   }
   else if(DS1961S)
   {
      if(owBuffer!=0x5F)
      {
         DEBUGLN("         Bad E/S");
         SuccessFlag = 0;
         return;
      }
      len = 8;
   }
   compareData = &Certificate.Record.ptr[offset];

   DEBUGBLOCKLBL("         Comparing this: ", compareData, len);
   DEBUG("                To this: ");
   do
   {
      ow_readbyte();
      DEBUGBLOCK(&owBuffer, 1);
      if(*compareData!=owBuffer)
      {
         DEBUGLN("");
         SuccessFlag = 0;
         return;
      }
      compareData++;
   }
   while(--len);

   DEBUGLN("");
   DEBUGLN("      read scratchpad back successful");
   SuccessFlag = 1;
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; The following COPY SP command CANNOT be retried in a DS1963S because the
; Page Write Counter could be incremented more than once. So this code ALWAYS
; returns good status no matter what really happens -
*/
void doCopyScratchpad63S(uchar page, uchar offset)
{
   // no retry because of write-cycle counter
   ow_select();
   if(SuccessFlag)
   {
      // copy scratchpad
      owBuffer = OW_COPY_SCRATCHPAD_CMD;
      ow_touchbyte();
      owBuffer = (page<<5) + offset;
      ow_touchbyte();
      owBuffer = page>>3;
      ow_touchbyte();
      owBuffer = 0x1F; // E/S byte
      ow_touchbyte();
      waitUntilDone();
   }
}

void doCopyScratchpad61S(uchar page, uchar offset)
{
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // copy scratchpad
         owBuffer = OW_COPY_SCRATCHPAD_CMD;
         ow_touchbyte();
         owBuffer = (page<<5) + offset;
         ow_touchbyte();
         owBuffer = page>>3;
         ow_touchbyte();
         owBuffer = (offset+7)&0x1F; // E/S byte
         ow_touchbyte();
         wait2ms();
         ow_touchblock(CopySPMAC, 20);
         wait10ms();
         ow_readbyte();
         ow_readbyte();
         ow_readbyte();
         //PRINTBLOCKLBL("         resp=", &ConfirmByte, 1);
         if(owBuffer==0xFF || owBuffer==0x00)
            SuccessFlag = 0;
         else
            SuccessFlag = 1;
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   //if(SuccessFlag)
   //{
      //PRINTLN("CopyScratchpad successful");
   //}
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	readBackVerify
;
; Desc:		After a Debit_Token has been performed, we must read back the token
;			and be sure that the debit worked. We perform a fully-authenticated
;			read-back to thwart play-back attacks.
;
; Exit:		SuccessFlag = True if success
*/
void readBackVerify(void)
{
   errorCode = 0;

   selectCoprocessor();
   getChallenge();
   if(!SuccessFlag)
   {
      errorCode = Err_GetChallenge;
      return;
   }

   selectCustomer();
   issueChallenge();
   if(!SuccessFlag)
   {
      errorCode = Err_IssueChallengeV;
      return;
   }

   readAccountPage();
   if(!SuccessFlag)
   {
      errorCode = Err_ReadAccount;
      return;
   }

   selectCoprocessor();
   authenticate();
   if(!SuccessFlag)
   {
      if(DS1961S)
         fixer();
      errorCode = Err_AuthenticateV;
      return;
   }

   // check to see if the file is what was supposed to be written
   //PRINTLN("   Check correct data:");
   //PRINTBLOCKLBL("      cur: ", Certificate.Record.ptr, 36);
   //PRINTBLOCKLBL("      new: ", NewRecord, 36);
   if(memcmp(Certificate.Record.ptr, NewRecord, 36)==0)
   {
      DEBUGLN("Successful Debit");
      SuccessFlag = 1;
      errorCode = Err_None;
      return;
   }

   // if it's a DS1961S and...
   if(DS1961S)
   {
      //PRINTBLOCKLBL("      Scr: ", ScratchRecord, 36);
      // if the first byte is all that differs from NewRecord
      // then the header wasn't updated
      if(memcmp(Certificate.Record.ptr+1, NewRecord+1, 35)==0)
      {
         // must attempt to re-write header
         memcpy(ScratchRecord, Certificate.Record.ptr, 36);
         memcpy(Certificate.Record.ptr, NewRecord, 36);
         selectCustomer();
         updateAccountPage61S(0);
         //if(SuccessFlag)
            //PRINTLN("Header updated successfully");
         SuccessFlag = 0;
         errorCode = Err_Tear_Reverify;
         return;
      }
   }

   //PRINTBLOCKLBL("      old: ", OldRecord, 36);
   // check to see if it's the same as the old data
   if(memcmp(Certificate.Record.ptr, OldRecord, 36)==0)
   {
      // must attempt to re-do entire update
      SuccessFlag = 0;
      errorCode = Err_Tear_Unchanged;
      return;
   }

   // attempt re-write
   //memcpy(Certificate.Record.ptr, NewRecord, 32);
   SuccessFlag = 0;
   errorCode = Err_Tear_Unknown;
   return;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: IncPageCounter
;
; Desc:		Increment the page write counter (as the iButton will) in prep
;			for re-signing the certificate
;
; Returns Carry True if the counter wrapped-around which means the device is done
*/
void incPageCounter(void)
{
   Certificate.WriteCounter[0]++;
   if(!Certificate.WriteCounter[0])
   {
      Certificate.WriteCounter[1]++;
      if(!Certificate.WriteCounter[1])
      {
         Certificate.WriteCounter[2]++;
         if(!Certificate.WriteCounter[2])
         {
            Certificate.WriteCounter[3]++;
         }
      }
   }
}



/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	ComputeFileCRC
;
; Desc:		Compute the appropriate inverted CRC16 value for the File record
;			and install it in the record.
*/
void computeFileCRC()
{
   crcl = CUST_ECP;
   crch = 0;
   // File Length is in the same position for both tokens
   crc16_block(Certificate.Record.ptr, Certificate.Record.rec.FileLength+1);
   Certificate.Record.ptr[Certificate.Record.rec.FileLength+1] = ~crcl;
   Certificate.Record.ptr[Certificate.Record.rec.FileLength+2] = ~crch;
}

void verifyFile()
{
   // File Length is in the same position for both tokens
   uchar FileLength = Certificate.Record.rec.FileLength;
   //PRINTBLOCK(Certificate.Record.ptr, 32);
   SuccessFlag = 1;
   if(DS1963S)
   {
      // Check the Account eCert File CRC and fixed values -
      if(FileLength!=29)
      {
         // wrong File length
         //PRINTLN("Bad Length");
         SuccessFlag = 0;
      }

      if(Certificate.Record.rec.Type!=1)
      {
         // wrong eCert type
         //PRINTLN("Bad Type");
         SuccessFlag = 0;
      }

      // TODO: Check the monetary units code, maybe
   }
   else if(DS1961S)
   {
      if(FileLength!=RECORD_A_LENGTH && FileLength!=RECORD_B_LENGTH)
      {
         // wrong File length
         //PRINTLN("Bad Length");
         SuccessFlag = 0;
      }

      if(Certificate.Record.rec33.Type!=1)
      {
         // wrong eCert type
         //PRINTLN("Bad Type");
         SuccessFlag = 0;
      }

      // TODO: Check the monetary units code, maybe
   }

   // check the file continuation ptr
   if(Certificate.Record.ptr[FileLength]!=0)
   {
      // wrong continuation pointer
      //PRINTLN("Bad CP");
      SuccessFlag = 0;
   }

   crcl = CUST_ECP;
   crch = 0;
   // File Length is in the same position for both tokens
   crc16_block(Certificate.Record.ptr, Certificate.Record.rec.FileLength+3);
   if(crcl!=0x01 || crch!=0xB0)
   {
      //PRINTLN("Bad File CRC16");
      SuccessFlag = 0;
   }
   //else
   //{
   //   PRINTLN("Good CRC16");
   //}


   if(DS1963S && !SuccessFlag)
   {
      return;
   }
   else if(DS1961S && !SuccessFlag)
   {
      // analyze A-B records to fix the file
      // save the original file length
      bit validA = 0, validB = 0;

      // save a copy of original, needed for update later
      memcpy(OldRecord, Certificate.Record.ptr, 32);
      memcpy(ScratchRecord, Certificate.Record.ptr, 32);

      // lets try Record A and check the crc
      Certificate.Record.rec.FileLength = RECORD_A_LENGTH;
      crcl = CUST_ECP;
      crch = 0;
      crc16_block(Certificate.Record.ptr, RECORD_A_LENGTH+3);
      if(crcl==0x01 && crch==0xB0)
      {
         validA = 1;
         //PRINTLN("Record A is valid");
      }

      // lets try Record B and check the crc
      Certificate.Record.rec.FileLength = RECORD_B_LENGTH;
      crcl = CUST_ECP;
      crch = 0;
      crc16_block(Certificate.Record.ptr, RECORD_B_LENGTH+3);
      if(crcl==0x01 && crch==0xB0)
      {
         validB = 1;
         //PRINTLN("Record B is valid");
      }

      if(validA && validB)
      {
         // Both A & B are valid!  And we know that we can only
         // get here if the pointer or the header was not valid.
         // That means that B was the last updated one but the
         // header got hosed and the debit was not finished...
         // which means A is the last known good value, let's go with A.
         Certificate.Record.rec.FileLength = RECORD_A_LENGTH;
         //PRINTLN("Updating account page to be Record A");
         writeAccountPage();
         //if(!SuccessFlag)
            //PRINTLN("Update failed");
      }
      else if(validA)
      {
         // B is invalid, A is valid.  Means A is the last updated one,
         // but B is the last known good value.  The header was not updated
         // to point to A before debit was aborted.  Let's go with B
         Certificate.Record.rec.FileLength = RECORD_B_LENGTH;
         // must fix B's CRC
         computeFileCRC();
         //PRINTLN("Updating account page to be Record B");
         writeAccountPage();
         //if(!SuccessFlag)
            //PRINTLN("Update failed");
      }
      else if(validB)
      {
         // A is invalid, B is valid. Should never ever happen.  Something
         // got completely hosed.  What should happen here?
         Certificate.Record.rec.FileLength = FileLength;
         //PRINTLN("No change is made");
      }
      else
      {
         // neither record contains a valid CRC.  What should happen here?
         Certificate.Record.rec.FileLength = FileLength;
         //PRINTLN("No change is made");
      }

      // regardless of whether or not data was fixed, return failure
      // and try the entire operation again.
      SuccessFlag = 0;
      return;
   }

   //PRINTLN("YEP, I'M HERE");
   //if(SuccessFlag)
   //   PRINTLN("IT'S TRUE");
   //else
   //   PRINTLN("IT'S FALSE");

   // Good file
   SuccessFlag = 1;
}

void makeNewTransID(uchar* TransID)
{
   //TransID[0] += 1;
   //if(TransID[0]==0)
   //   TransID[1] += 1;

   // cheezy randomizer.
   TransID[0] ^= (Entropy&0x00FF);
   TransID[1] ^= (Entropy&0xFF00)>>8;
}


/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	getCopyScratchpadMAC
;
; Desc:		When a DS1961S will be written-to, a MAC must be used to access
;			the device (to perform the Copy SP command). This builds the MAC.
;
; Entry:	OldRecord contains the original data page
;			eCert contains the new data that WILL be in the SP
;
; Note:		When a DS1961S is used, the co-processor's monetary validation
;			secret (for use with DS1963S devices) is obliterated. The system
;			can never function with DS1963S devices again until the co-processor
;			is regenerated. This is done in ComputeDeviceSecret if the device
;			if a DS1961S.
*/
void getCopyScratchpadMAC(uchar offset)
{
   // ensure access to the scratchpad
   eraseScratchpad();
   if(!SuccessFlag)
      return;

   //PRINTBLOCKLBL("gCSMAC:     OldRecord: ", OldRecord, 36);
   //PRINTBLOCKLBL("gCSMAC: ScratchRecord: ", ScratchRecord, 36);
   //PRINTBLOCKLBL("gCSMAC: offset: ", &offset, 1);
   //PRINTBLOCKLBL("gCSMAC: &Certificate.Record.ptr[offset]: ", &Certificate.Record.ptr[offset], 8);

   // write the first 28 bytes of old page data, followed by four bytes of
   // new page data to the signing data page (usually page 8).
   // write the data first to the scratchpad
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write Scratchpad command
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_SDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         ow_touchblock_crc(ScratchRecord, 28);
         ow_touchblock_crc(&Certificate.Record.ptr[offset], 4);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   // Copy the SP data to the page for signing -
   Retry = RetryCount;
   do
   {
      doCopyScratchpad63S(COP_SDP, 0);
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   // Send the second four bytes of new data to scratchpad -
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write Scratchpad command
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         owBuffer = (COP_SDP<<5) + 8; // Set offset of scratchpad to 8
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         // the second group of four new data bytes
         ow_touchblock_crc(&Certificate.Record.ptr[offset+4], 4);
         // send the page # and romID(-crc8)
         ow_touchblock_crc(&Certificate.PageNumber, 8);
         sendNulls(12);
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   // Perform Sign Data Page command -
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Start a SHA operation;
         owBuffer = OW_COMPUTE_SHA_CMD;
         ow_touchbyte_crc();
         owBuffer = COP_SDP<<5;
         ow_touchbyte_crc();
         owBuffer = COP_SDP>>3;
         ow_touchbyte_crc();
         // Perform the Sign Data Page operation
         owBuffer = OW_SIGN_DATA_PAGE_CMD;
         ow_touchbyte_crc();
         verifyCRC();
         if(SuccessFlag)
            waitUntilDone();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   if(!SuccessFlag)
      return;

   // do a dummy write to set offset of scratchpad to 8
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Write Scratchpad command
         owBuffer = OW_WRITE_SCRATCHPAD_CMD;
         ow_touchbyte();
         owBuffer = 8; // Set offset of scratchpad to 8
         ow_touchbyte();
         owBuffer = 0;
         ow_touchbyte();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   // read the scratchpad and store the signature
   Retry = RetryCount;
   do
   {
      ow_select();
      if(SuccessFlag)
      {
         // Read Scratchpad operation
         owBuffer = OW_READ_SCRATCHPAD_CMD;
         ow_touchbyte_crc();
         sendNulls(3); // Discard TA1, TA2, and E/S
         ow_readblock_crc(CopySPMAC, 20);
         sendNulls(4); // go to end of scratchpad
         verifyCRC();
      }
      if(SuccessFlag)
         break;
      NeedAccess = 1;
   }
   while(--Retry);

   DEBUGBLOCKLBL("   Copy Scratchpad MAC: ", CopySPMAC, 20);
}

// refreshes the entire account page for DS1961S,
// fixing all weakly programmed bits
void fixer(void)
{
   uchar offset;
   SuccessFlag = 1;
   for(offset=0; offset<32; offset+=8)
   {
      Retry = 2;
      do
      {
         ow_select();
         if(SuccessFlag)
         {
            owBuffer = OW_REFRESH_SCRATCHPAD_CMD;
            ow_touchbyte_crc();
            owBuffer = (CUST_ECP<<5)+offset;
            ow_touchbyte_crc();
            owBuffer = CUST_ECP>>3;
            ow_touchbyte_crc();
            sendNulls(8);
            verifyCRC();
         }
         if(SuccessFlag)
         {
            ow_select();
            if(SuccessFlag)
            {
               owBuffer = OW_LOAD_FIRST_SECRET_CMD;
               ow_touchbyte();
               owBuffer = (CUST_ECP<<5)+offset;
               ow_touchbyte();
               owBuffer = CUST_ECP>>3;
               ow_touchbyte();
               owBuffer = (offset+7)&0x1F; // E/S byte
               ow_touchbyte();
               wait10ms();
               ow_readbyte();
               ow_readbyte();
               ow_readbyte();
               if(owBuffer!=0xFF && owBuffer!=0x00)
                  SuccessFlag = 1;
               else
                  SuccessFlag = 0;
            }
         }
         NeedAccess = 1;
      }
      while(!SuccessFlag && --Retry);
   }
   //if(SuccessFlag)
   //   PRINTLN("   Fixer Success");
   //else
   //   PRINTLN("   Fixer Failed");
}

void sendNulls(uchar len)
{
   dbyte cnt = len;
   do
   {
      ow_readbyte_crc();
   }
   while(--cnt);
}
