// file ecow.c
#pragma src(ecow.a51)

#include "ecash.h"

#define DOOWDEBUG 0

#if OSCILLATOR==33000000
   #define FOUR_HUNDRED_FIFTY_NS(value) {owDelayCounter = value*3; while(--owDelayCounter);}
   #define DELAY_BEFORE_BIT() FOUR_HUNDRED_FIFTY_NS(1)
   #define DELAY_BEFORE_SAMPLE() FOUR_HUNDRED_FIFTY_NS(2)
   #define DELAY_BEFORE_TIMESLOT_END() FOUR_HUNDRED_FIFTY_NS(7)
   #define DELAY_FOR_RECOVERY() FOUR_HUNDRED_FIFTY_NS(5)
#elif OSCILLATOR==22118400
   #define FOUR_HUNDRED_FIFTY_NS(value) {owDelayCounter = value*2; while(--owDelayCounter);}
   #define DELAY_BEFORE_BIT() FOUR_HUNDRED_FIFTY_NS(1)
   #define DELAY_BEFORE_SAMPLE() FOUR_HUNDRED_FIFTY_NS(1)
   #define DELAY_BEFORE_TIMESLOT_END() FOUR_HUNDRED_FIFTY_NS(7)
   #define DELAY_FOR_RECOVERY() FOUR_HUNDRED_FIFTY_NS(5)
#elif OSCILLATOR==11059200
   #define FOUR_HUNDRED_FIFTY_NS(value) {owDelayCounter = value; while(--owDelayCounter);}
   #define DELAY_BEFORE_BIT() {owDelayCounter = 0;}
   #define DELAY_BEFORE_SAMPLE() FOUR_HUNDRED_FIFTY_NS(1)
   #define DELAY_BEFORE_TIMESLOT_END() FOUR_HUNDRED_FIFTY_NS(7)
   #define DELAY_FOR_RECOVERY() {owDelayCounter = 1; owDelayCounter = 0;}
#endif

// loop counter variables
static dbyte owCounter, owDelayCounter;
//dint COPR_STD_Bit, CUST_STD_Bit, COPR_OD_Bit, CUST_OD_Bit;
//dint COPR_STD_Reset, CUST_STD_Reset, COPR_OD_Reset, CUST_OD_Reset;


bbyte owBuffer; // buffer used for touchbyte/touchbit routines
DECLARE_BIT(owBuffer, 0, owBufferOut); // LSB of Buffer
DECLARE_BIT(owBuffer, 7, owBufferIn); // MSB of buffer

DECLARE_BIT(P3, 4, OW_AUXBus); // Auxiliary 1-Wire Internal Clock/RAM Bus

// Coprocessor Port Pins
DECLARE_BIT(P1, 4, OW_COPSense);
DECLARE_BIT(P1, 5, OW_COPDrive); // (Inverted by FET, 1=Drive Bus Low)

// Customer Port Pins -
DECLARE_BIT(P1, 6, OW_CUSTSense);
DECLARE_BIT(P1, 7, OW_CUSTDrive); // (Inverted by FET, 1=Drive Bus Low)

void ow_init(void)
{
   OW_COPDrive = 0; // lower (kill) drives
   OW_COPSense = 1;
   OW_CUSTDrive = 0; // lower (kill) drives
   OW_CUSTSense = 1;

   // delay before first reset, to power up parts
   msDelay(1);
}

/*
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: OW_Select
;
; Desc:     Address the part named in RomData on the current port. If the
;           NeedAccess flag is set, then the part is first reset at standard
;           speed and placed into Overdrive mode. Resume is used in normal
;           mode, and Match ROM (RomData) is used in NeedAccess modes.
*/
void ow_select(void)
{
   if(NeedAccess)
   {
      OWOverdrive = 0; // slow to standard speed
      ow_reset();
      if(!SuccessFlag) // if no presence pulse
         return;
      if(Mode_ODEnabled)
      {
         owBuffer = 0x69; // Overdrive Match ROM command
         ow_touchbyte();
         OWOverdrive = 1; // back to overdrive speed
      }
      else
      {
         owBuffer = 0x55; // Match ROM command
         ow_touchbyte();
      }
      // send the address
      owBuffer = RomData[0];
      #pragma asm
         mov B, #8
      ow_select_loop_1:
         call ow_touchbyte
         inc dptr
         movx A, @dptr
         mov owBuffer, A
         djnz B, ow_select_loop_1
      #pragma endasm
      NeedAccess = 0;
   }
   else // Resume
   {
      ow_reset();
      if(!SuccessFlag) // if no presence pulse
         return;
      owBuffer = 0xA5; // Resume command
      ow_touchbyte();
   }
   crcl = 0;
   crch = 0;
   SuccessFlag = 1;
}

// --------------------------------------------------------------------
// ow_reset()
//
// Resets the selected 1-wire port.
//
// COPSelected bit indicates coprocessor port, otherwise customer
// OWOverdrive bit indicates overdrive reset, otherwise standard speed
// --------------------------------------------------------------------
void ow_reset(void)
{
   if(OWOverdrive) // overdrive reset
   {
      // need 5us recovery time before a reset
      // all operations leave 3us, so this should cover it
      usDelay(2);
      if(COPSelected) // overdrive reset coprocessor port
      {
         OW_COPDrive = 1; // Start a reset pulse
         // delay for 60us - was 48us, DS1961S prefers 60us
         usDelay(60);

         EA = 0; // disable all interrupts
         OW_COPDrive = 0; // end the reset pulse

         usDelay(1); // wait for rise time

         // make sure it isn't shorted
         if(OW_COPSense)
         {
            owCounter = 8;
            do
            {
               if(!OW_COPSense)
               {
                  EA = 1; // re-enable interrupts
                  while(!OW_COPSense)
                     ;  // wait for the end of the presence pulse(s)
                  usDelay(2); // recovery time
                  SuccessFlag = 1; // signal that we got a presence pulse
                  return;
               }
               usDelay(1);
            }
            while(--owCounter);
         }
         EA = 1; // re-enable interrupts
         SuccessFlag = 0; // signal that there were no presence pulses
      }
      else // overdrive reset customer port
      {
         OW_CUSTDrive = 1; // Start a reset pulse
         // delay for 60us - was 48us, DS1961S prefers 60us
         usDelay(60);

         EA = 0; // disable all interrupts
         OW_CUSTDrive = 0; // end the reset pulse

         usDelay(1); // wait for rise time

         // make sure it isn't shorted
         if(OW_CUSTSense)
         {
            owCounter = 8;
            do
            {
               if(!OW_CUSTSense)
               {
                  EA = 1; // re-enable interrupts
                  while(!OW_CUSTSense)
                     ;  // wait for the end of the presence pulse(s)
                  usDelay(2); // recovery time
                  SuccessFlag = 1; // signal that we got a presence pulse
                  return;
               }
               usDelay(1);
            }
            while(--owCounter);
         }
         EA = 1; // re-enable interrupts
         SuccessFlag = 0; // signal that there were no presence pulses
      }
   }
   else // Standard speed reset
   {
      if(COPSelected) // standard speed reset coprocessor
      {
         OW_COPDrive = 1; // Start a reset pulse

         // generate 480us delay
         usDelay(480);

         EA = 0; // disable all interrupts
         OW_COPDrive = 0; // end the reset pulse

         usDelay(3); // wait for rise time

         if(OW_COPSense) // if line is not shorted
         {
            owCounter = 70;
            do
            {
               if(!OW_COPSense) // if presence pulse occurs
               {
                  EA = 1; // re-enable interrupts
                  while(!OW_COPSense)
                     ; // Wait for end of presence pulse(s)
                  usDelay(5);
                  SuccessFlag = 1; // signal that we got a presence pulse
                  return;
               }
               usDelay(1);
            }
            while(--owCounter);
         }

         EA = 1; // re-enable interrupts

         usDelay(240);

         SuccessFlag = 0; // Signal bad if no presence pulses found
                          // or shorted line...
      }
      else // standard speed reset customer
      {
         OW_CUSTDrive = 1; // Start a reset pulse

         // generate 480us delay
         usDelay(480);

         EA = 0; // disable all interrupts
         OW_CUSTDrive = 0; // end the reset pulse

         usDelay(3); // wait for rise time

         if(OW_CUSTSense) // if line is not shorted
         {
            owCounter = 70;
            do
            {
               if(!OW_CUSTSense) // if presence pulse occurs
               {
                  EA = 1; // re-enable interrupts
                  //puts("Found a part");
                  while(!OW_CUSTSense)
                     ; // Wait for end of presence pulse(s)
                  usDelay(5);
                  SuccessFlag = 1; // signal that we got a presence pulse
                  return;
               }
               usDelay(1);
            }
            while(--owCounter);
         }

         EA = 1; // re-enable interrupts

         usDelay(240);

         SuccessFlag = 0; // Signal bad if no presence pulses found
                          // or shorted line...
      }
   }
}

void ow_readbyte(void)
{
   owBuffer = 0xFF;
   ow_touchbyte();
}

void ow_readbyte_crc(void)
{
   owBuffer = 0xFF;
   ow_touchbyte();
   crc16_byte(owBuffer);
}


void ow_readblock(uchar* block, uchar len)
{
   owCounter = len;
   owBuffer = block[0]; // just to set dptr to address of 'block'
   do
   {
      owBuffer = 0xFF;
      ow_touchbyte();
      #pragma asm
         mov A, owBuffer
         movx @dptr, A
         inc dptr
      #pragma endasm
   }
   while(--owCounter);
}


void ow_readblock_crc(uchar* block, uchar len)
{
   owCounter = len;
   owBuffer = block[0]; // just to set dptr to address of 'block'
   do
   {
      owBuffer = 0xFF;
      ow_touchbyte();
      crc16_byte(owBuffer);
      #pragma asm
         mov A, owBuffer
         movx @dptr, A
         inc dptr
      #pragma endasm
   }
   while(--owCounter);
}

// does a standard speed read of the romID on the selected port
// including crc8
void ow_readRomID(void)
{
   // slow to standard speed
   OWOverdrive = 0;
   ow_reset();
   if(!SuccessFlag)
   {
      return;
   }

   if(Mode_ODEnabled)
   {
      // Overdrive Skip ROM
      owBuffer = OW_OD_SKIP_ROM_CMD;
      ow_touchbyte();

      OWOverdrive = 1;
      ow_reset();
      if(!SuccessFlag)
      {
         return;
      }
   }

   owBuffer = 0x33; // Read ROM command
   ow_touchbyte();

   crcl = 0; // reset the crc8
   owCounter = 8;

   owBuffer = RomData[0]; // set dptr to point to block
   do
   {
      ow_readbyte();
      #pragma asm
         mov A, owBuffer
         movx @dptr, A  ; Read and store 8 ROM ID bytes
         inc dptr
      #pragma endasm

      crc8_byte(owBuffer); // check the crc8 as we go
   }
   while(--owCounter);

   // check the crc
   if(crcl!=0)
   {
      SuccessFlag = 0;
      return;
   }

   // if it matches, let's Reset-MatchROM, so the part is selected
   ow_reset();
   if(!SuccessFlag) // if no presence pulse
      return;
   owBuffer = 0x55; // Match ROM command
   ow_touchbyte();
   // send the address
   owBuffer = RomData[0];
   #pragma asm
      mov B, #8
   ow_readRomID_select_loop_1:
      call ow_touchbyte
      inc dptr
      movx A, @dptr
      mov owBuffer, A
      djnz B, ow_readRomID_select_loop_1
   #pragma endasm
   NeedAccess = 0;
}

void ow_touchbyte(void)
{
   //owBuffer = ACC;
   ow_touchbit();
   ow_touchbit();
   ow_touchbit();
   ow_touchbit();
   ow_touchbit();
   ow_touchbit();
   ow_touchbit();
   ow_touchbit();
   //ACC = owBuffer;
}

void ow_touchblock(uchar* block, uchar len)
{
   owCounter = len;
   owBuffer = block[0]; // set dptr to point to block
   do
   {
      ow_touchbyte();
      #pragma asm
         inc dptr
         movx A, @dptr
         mov owBuffer, A
      #pragma endasm
   }
   while(--owCounter);
}

void ow_touchblock_crc(uchar* block, uchar len)
{
   owCounter = len;
   owBuffer = block[0]; // set dptr to point to block
   do
   {
      ow_touchbyte();
      crc16_byte(owBuffer);
      #pragma asm
         inc dptr
         movx A, @dptr
         mov owBuffer, A
      #pragma endasm
   }
   while(--owCounter);
}


void ow_touchbyte_crc(void)
{
   ow_touchbyte();
   crc16_byte(owBuffer);
}

// --------------------------------------------------------------------
// ow_touchbit()
//
// Does a single bit of communication on the 1-Wire port.
//
// COPSelected bit indicates coprocessor port, otherwise customer
// OWOverdrive bit indicates overdrive reset, otherwise standard speed
//
// Standard Speed:
//    Start Time Slot       @0 us
//    Write-1 Set Bit High  @5 us
//    Read Sample           @15 us
//    End Time Slot         @60 us
//    End Bit (w/ Recovery) @65 us
// Throughput = 15 kbps
//
// Overdrive Speed:
//    Start Time Slot       @0 us
//    Write-1 Set Bit High  @1 us
//    Read Sample           @2 us
//    End Time Slot         @6 us
//    End Bit (w/ Recovery) @8 us
// Throughput = 125 kbps
//
// --------------------------------------------------------------------
void ow_touchbit(void)
{
   EA = 0;
   if(OWOverdrive) // overdrive speed
   {
      // overdrive touch bit coprocessor
      if(COPSelected)
      {
         //COPR_OD_Bit++;
         // start timeslot.
         OW_COPDrive = 1;
         DELAY_BEFORE_BIT();
         // send bit out.
         OW_COPDrive = !owBufferOut;
         DELAY_BEFORE_SAMPLE();
         // sample result
         owBuffer >>= 1;
         owBufferIn = OW_COPSense;
         DELAY_BEFORE_TIMESLOT_END();
         // end timeslot.
         OW_COPDrive = 0;
         DELAY_FOR_RECOVERY();
         EA = 1;
         return;
      }
      else // if customer port selected
      {
         //CUST_OD_Bit++;
         // start timeslot.
         OW_CUSTDrive = 1;
         DELAY_BEFORE_BIT();
         // send bit out.
         OW_CUSTDrive = !owBufferOut;
         DELAY_BEFORE_SAMPLE();
         // sample result
         owBuffer >>= 1;
         owBufferIn = OW_CUSTSense;
         DELAY_BEFORE_TIMESLOT_END();
         // timeslot done.
         OW_CUSTDrive = 0;
         DELAY_FOR_RECOVERY();
         EA = 1;
         return;
      }
   }
   else //standard speed touch bit
   {
      // standard speed touch bit coprocessor
      if(COPSelected)
      {
         //COPR_STD_Bit++;
         // start timeslot.
         OW_COPDrive = 1;
         usDelay(5);
         // send bit out.
         OW_COPDrive = !owBufferOut;
         usDelay(10);
         // sample result @ 15 us.
         owBuffer >>= 1;
         owBufferIn = OW_COPSense;
         usDelay(40);
         // end timeslot.
         OW_COPDrive = 0;
         usDelay(3);
         EA = 1;
         return;
      }
      else // if customer port selected
      {
         //CUST_STD_Bit++;
         // start timeslot
         OW_CUSTDrive = 1;
         usDelay(5);
         // send bit out.
         OW_CUSTDrive = !owBufferOut;
         usDelay(10);
         // sample result @ 15 us.
         owBuffer >>= 1;
         owBufferIn = OW_CUSTSense;
         usDelay(40);
         // end timeslot.
         OW_CUSTDrive = 0;
         usDelay(3);
         EA = 1;
         return;
      }
   }
}


/*
void bit_functions(void)
{
#pragma asm
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; CPU is DS5000 at 14.7456 MHz, cycle time is 814ns.

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: COP_OD_TBit
;
; Desc:     Co-Processor Overdrive Speed Touch Bit
;
; Performance:
;       Start pulse     0.814 us (1 cycle)
;       Sample          2.035 us (2.5 cycles)
;       Write Zero      7.324 us (9 cycles)
;       Bit Time        9.766 us (12 cycles)
;       Bit Rate        102 KBPS
COP_OD_TBit:

    clr EA
    call COP_OD_TBix
    nop                         ; 1     7
    nop                         ; 1     8
    clr OW_COPDrive             ;       9
    setb EA
    ret

COP_OD_TBix:

    clr OW_COPDrive             ; 1     9
    jb ACC.0, COP_OD_TBix_1     ; 2     10
        setb OW_COPDrive        ; 1     0/12
        clr C                   ; 1     1
    jmp COP_OD_TBix_end         ; 2     2
COP_OD_TBix_1:
        setb OW_COPDrive        ; 1         0
        clr OW_COPDrive         ; 1         1
        nop                     ; 1         2
        mov C,OW_COPSense       ; 1         3
COP_OD_TBix_end:                ; 0     4   4
    rrc A                       ; 1     4
    ret                         ; 2     5
;   call COP_OD_TBix            ; 2     7


;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: COP_OD_TByte
;
; Desc:     Co-Processor Overdrive Speed Touch Byte, Read Byte
;
COP_OD_RByte:

    mov A,#0FFH

COP_OD_TByte:

    clr EA
    call COP_OD_TBix            ; Perform a Touch Bit 1
    call COP_OD_TBix            ; Perform a Touch Bit 2
    call COP_OD_TBix            ; Perform a Touch Bit 3
    call COP_OD_TBix            ; Perform a Touch Bit 4
    call COP_OD_TBix            ; Perform a Touch Bit 5
    call COP_OD_TBix            ; Perform a Touch Bit 6
    call COP_OD_TBix            ; Perform a Touch Bit 7
    call COP_OD_TBix            ; Perform a Touch Bit 8
    nop                         ; 1     7
    nop                         ; 1     8
    clr OW_COPDrive             ; 1     9
    setb EA
    ret

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: COP_Std_TBit
;
; Desc:     Co-Processor Standard Speed Touch Bit
;
COP_Std_TBit:

    clr EA
    setb OW_COPDrive            ; 1     0/79        Start the time slot
    jnb ACC.0,COP_Std_TBit_0    ; 2     1
        nop                     ; 1     3-------+
        nop                     ; 1     4       |
        nop                     ; 1     5       |
        clr OW_COPDrive         ; 1     6       |   End the Write One pulse (4.883 us)
        mov R0,#4               ; 1     7       |
    jmp COP_Std_TBit_End        ; 2     8       |
COP_Std_TBit_0:
        mov R0,#7               ; 1             3
COP_Std_TBit_End:               ; 0     10      4
    djnz R0, $                  ; 2*R0  10      4
    mov C,OW_COPSense           ; 1     18      18  Sample (14.648 us)
    rrc A                       ; 1     19
    mov R0,#26                  ; 1     20
    djnz R0, $                  ; 2*R0  21
    clr OW_COPDrive             ; 1     73          End the write Zero pulse (59.408 us)
    setb EA                     ; 1     74
    ret                         ; 2     75          Release = 6 (4.883 us)
;   call COP_Std_TBit           ; 2     77          Bit Rate = 15,554 BPS



;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: COP_Std_TByte
;
; Desc:     Co-Processor Standard Speed Touch Byte
;
COP_Std_RByte:

    mov A,#0FFH

COP_Std_TByte:

    call COP_Std_TBit           ; Bit 0
    call COP_Std_TBit           ; Bit 1
    call COP_Std_TBit           ; Bit 2
    call COP_Std_TBit           ; Bit 3
    call COP_Std_TBit           ; Bit 4
    call COP_Std_TBit           ; Bit 5
    call COP_Std_TBit           ; Bit 6
    call COP_Std_TBit           ; Bit 7
    ret

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: CUST_OD_TBit
;
; Desc:		Customer Overdrive Speed Touch Bit
;
; Performance:
;		Start pulse		0.814 us (1 cycle)
;		Sample			2.035 us (2.5 cycles)
;		Write Zero		7.324 us (9 cycles)
;		Bit Time		9.766 us (12 cycles)
;		Bit Rate		102 KBPS
CUST_OD_TBit:
	clr EA
	call CUST_OD_TBix
	clr OW_CUSTDrive
	nop							; 1		7
	nop							; 1		8
	setb EA						;		9
	ret

CUST_OD_TBix:
	clr OW_CUSTDrive			; 1		9
	jb ACC.0, CUST_OD_TBix_1    ; 2		10
		setb OW_CUSTDrive		; 1		0/12
		clr C					; 1		1
	jmp CUST_OD_TBix_End        ; 2		2
CUST_OD_TBix_1:
		setb OW_CUSTDrive		; 1			0
		clr OW_CUSTDrive		; 1			1
		nop						; 1			2
		mov C,OW_CUSTSense		; 1			3
CUST_OD_TBix_End:				; 0		4	4
	rrc A						; 1		4
	ret							; 2		5
;	call CUST_OD_TBix			; 2		7



;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: CUST_OD_TByte
; Desc:		Customer Overdrive Speed Touch Byte, Read byte

CUST_OD_RByte:

	mov A,#0FFH

CUST_OD_TByte:

	clr EA
	call CUST_OD_TBix
	call CUST_OD_TBix
	call CUST_OD_TBix
	call CUST_OD_TBix
	call CUST_OD_TBix
	call CUST_OD_TBix
	call CUST_OD_TBix
	call CUST_OD_TBix
	nop							; 1		7
	nop							; 1		8
	clr OW_CUSTDrive			; 1		9
	setb EA
	ret


;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

; Function: CUST_Std_TBit

; Desc:		Customer Standard Speed Touch Bit

CUST_Std_TBit:

	clr EA
	setb OW_CUSTDrive			; 1		0/79		Start the time slot
	jnb ACC.0, CUST_Std_TBit_0  ; 2		1
		nop						; 1		3-------+
		nop						; 1		4       |
		nop						; 1		5		|
		clr OW_CUSTDrive		; 1		6		|	End the Write One pulse (4.883 us)
		mov R0,#4				; 1		7		|
	jmp CUST_Std_Tbit_End		; 2		8		|
CUST_Std_TBit_0:
		mov R0,#7				; 1				3
CUST_Std_Tbit_End:				; 0		10		4
	djnz R0,$       			; 2*R0	10		4
	mov C,OW_CUSTSense			; 1		18		18	Sample (14.648 us)
	rrc A						; 1		19
	mov R0,#26					; 1		20
	djnz R0,$       			; 2*R0	21
	clr OW_CUSTDrive			; 1		73			End the write Zero pulse (59.408 us)
	setb EA						; 1		74
	ret							; 2		75			Release = 6 (4.883 us)
;	call CUST_Std_TBit			; 2		77			Bit Rate = 15,554 BPS



;+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

; Function: CUST_Std_TByte

; Desc:		Customer Standard Speed Touch Byte, Read Byte

CUST_Std_RByte:
	mov A,#0FFH

CUST_Std_TByte:
	call CUST_Std_TBit
	call CUST_Std_TBit
	call CUST_Std_TBit
	call CUST_Std_TBit
	call CUST_Std_TBit
	call CUST_Std_TBit
	call CUST_Std_TBit
	call CUST_Std_TBit
	ret
#pragma endasm
}
*/