// ectimer.c
#include "ecash.h"

/*
    mov TMOD,#00000001B         ; Timer 0 to Mode 1
    mov TCON,#00001000B         ; Timer 0 to RUN mode
 %IF Processor eq 520
    anl CKCON,#11110111B        ; Timer 0 to Divide-by-12 mode
 %ENDIF
    setb ET0                    ; Enable the Timer 0 Interrupt
    setb TR0
*/
void timer0Init(void)
{
   TMOD = 0x01; // Timer0 to Mode 1
   TCON = 0x08; // Timer0 to RUN mode
   // if 520, anl CKCON, #11110111B; Timer 0 to Divide-by-12 mode
   // 420 defaults to div-by-12 mode

   QtrSeconds = 4;
   milliseconds = 250;
   TH0 = ((TIMER0_1MSEC_RELOAD)>>8)&0x0FF;
   TL0 = (TIMER0_1MSEC_RELOAD)&0x0FF;

   ET0 = 1;
   TR0 = 1;
   EA = 1;
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function: Int_Timer0
;
; Desc:     Process a Timer 0 interrupt. Timer 0 is running at all times and
;           is used for general system timing.  It is pre-scaled to a 1kHz
;           tick rate.
;
; Setup:    Timer in Mode 1 (16 bit) (TMOD = XXXX0001B)
;           Pre-scaler to Divide-by-12 (CKCON.3 = 0)
;           Timer 0 Interrupt Enabled
;           Timer Run (TR0) = High
;
*/
void timer0(void) interrupt 1
{
   TH0 = ((TIMER0_1MSEC_RELOAD)>>8)&0x0FF;
   TL0 = (TIMER0_1MSEC_RELOAD)&0x0FF;
   if(!(--milliseconds))
   {
      milliseconds = 250;
      if(!(--QtrSeconds))
      {
         QtrSeconds = 4;
         SecBit = !SecBit;
         if(secTimerRun)
         {
            if(!(--TmrSeconds))
               TimeOut = 1;
         }
      }
   }

   if(msTimerRun)
   {
      msCounter++;
      if(msCounter==0)
      {
         msTimerOverflow = 1;
      }
   }
}

void startTimerSec()
{
   // stop the timer
   secTimerRun = 0;

   // load the timeout value
   TmrSeconds = TimeOutValue;

   // clear the timeout flag
   TimeOut = 0;

   // start the second timer
   secTimerRun = 1;
}

void stopTimerSec(void)
{
   // stop the timer
   secTimerRun = 0;

   // clear the timeout flag
   TimeOut = 0;
}

void startTimerMS(void)
{
   // stop the timer
   msTimerRun=0;

   // reset the counter
   msCounter=0;

   // clear the overflow flag
   msTimerOverflow=0;

   // start the timer
   msTimerRun=1;
}

void stopTimerMS(void)
{
   msTimerRun=0;
   //if(msTimerOverflow)
   //   PRINTLN("= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = Time Overflowed");
   //else
   //   PRINTBLOCKLBL("= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = Time (hex ms)=", &msCounter, 1);
   //msTimerOverflow = 0;
}

// generates a microsecond delay of the specified amount of time
void usDelay(unsigned int delay)
{
#if OSCILLATOR==33000000
   dbyte tempCounter;
   do
   {
      tempCounter = 4; // = 3 cycles
      while(--tempCounter); // = 5 cycles * 4 = 20 cycles
   }
   while(--delay); // = 10/11 cycles
   // total = (3+20+[10/11]) = [33/34] cycles
   // total = [.99/1.20] microseconds
#elif OSCILLATOR==22118400
   dbyte tempCounter;
   do
   {
      tempCounter = 2; // = 3 cycles
      while(--tempCounter); // = 5 cycles * 2 = 10 cycles
   }
   while(--delay); // = 10/11 cycles
   // total = (3+10+[10/11]) = [23/24] cycles
   // total = [1.04/1.08] microseconds
#elif OSCILLATOR==11059200
   do
   {
      ACC = 0; // 1 cycle
   }
   while(--delay); // = 10 cycles
   // total = 11 cycles = 1.10 microseconds
#endif
}

void msDelay(unsigned int delay)
{
   do
   {
      usDelay(1000);
   }
   while(--delay);
}

/*
;++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
; Function:	Wait2ms
;
; Desc:		Delay for 2 ms
;
; Note:		Xtal is 14.7456 MHz, cycle time depends on CPU type.
;			DS5000 (/12) = 814 ns, DS87C520 (/4) = 271 ns.
;
; DS89C420 (/1) @ 11.095MHz = 90 ns, 2ms = 22,222 instruction cycles
; DS89C420 (/1) @ 22.190MHz = 45 ns, 2ms = 44,444 instruction cycles
; DS89C420 (/1) @ 33.000MHz = 30 ns, 2ms = 66,666 instruction cycles
*/
void wait2ms(void)
{
   msDelay(2);
}

void wait10ms(void)
{
   msDelay(10);
}

void wait250ms(void)
{
   msDelay(250);
}





void watchdogInit()
{
   //RWT = 1;
}

/*
Int_WatchDog:
!1  TAccess
!1  setb RWT
!1  mov dptr,#Msg_Watchdog!!!!
!1  call LCD_WriteMessage
    jmp 0

void watchdog_interrupt()
{
   TAccess
   RWT = 1;
   LCD_WriteMessag(Msg_Watchdog);
}*/