import com.dalsemi.onewire.utils.Convert;
import com.dalsemi.onewire.adapter.SerialService;
import java.util.*;
import java.io.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import gnu.io.*;
//import javax.comm.*;

public class eCashMonitor
   extends JFrame
   implements ActionListener, SerialPortEventListener
{
   private static final Font font = new Font("DialogInput", Font.PLAIN, 12);

   public boolean WINDOWS    = false;
   public boolean MACOSX     = false;

   //Default values
   protected int     height          = 550;
   protected int     width           = 610;
   protected int     xPos            = 0;
   protected int     yPos            = 0;
   protected String  baud            = "57600";
   protected String  lastUsedDir     = null;

   private SerialService serialService = null;

   private JTextArea readText;
   private JButton openCloseButton;
   private JButton resetButton;
   private JRadioButton dtrSet;
   private JRadioButton dtrClear;

   private JMenu fileMenu;
      private JMenuItem exitItem;
   private JMenu editMenu;
   private JMenu loaderMenu;
      private JMenuItem loadHexFileItem;
      private JMenuItem verifyHexFileItem;
   private JMenu optionsMenu;
      private JMenuItem showRawMenuItem;
      private JMenu setMenu;
         private JMenuItem setServiceFilename;
         private JMenuItem setLCDText;
      private JMenu getMenu;
   private JPopupMenu rightClickPopupMenu;
   private JLabel statusLabel;

   private JMenuBar menuBar = new JMenuBar();

   private JComboBox portChoice;
   private JComboBox baudChoice;

   private boolean showRawMessages = false;
   private boolean portOpened = false;

   JFileChooser chooser;

   public eCashMonitor(String strComPort, String strBaud)
      throws IOException
   {
      super("eCash Monitor");
      
      String os = System.getProperty("os.name");
      if ((os != null) && (os.indexOf("Windows") == -1))
      {
         WINDOWS = false;
         if (os.startsWith("Mac OS X"))
            MACOSX = true;
      }
      else
      {
         WINDOWS = true;
      }

      Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
      xPos = screenSize.width/2 - width/2;
      yPos = screenSize.height/2 - height/2;
      loadINIFile();
      setLocation(xPos,yPos);
      setSize(width, height);

      chooser = new JFileChooser(lastUsedDir);
      chooser.setMultiSelectionEnabled(true);
      chooser.resetChoosableFileFilters();
      chooser.setFileFilter(new HexFileFilter());

      if(strComPort!=null)
      {
         try
         {
            serialService = SerialService.getSerialService(strComPort);
            serialService.openPort((SerialPortEventListener)this);
            if(strBaud!=null)
              serialService.setBaudRate(Integer.parseInt(strBaud));
         }
         catch(Exception e)
         {
            strComPort = null;
            strBaud = null;
         }
      }

      Container content = this.getContentPane();
      content.setLayout(new BorderLayout());
      content.setFont(font);
      this.addWindowListener(new WindowAdapter()
         {
            public void windowClosing(WindowEvent we)
            {
               exitApplication();
            }
         }
      );

      JPanel southPanel = new JPanel();
      southPanel.setLayout(new BorderLayout());
         JPanel configurationPanel = new JPanel();
         configurationPanel.setLayout(new GridLayout(2, 2));
            JLabel portNameLabel = new JLabel("Port Name:", Label.LEFT);
            portNameLabel.setFont(font);
            JLabel baudLabel = new JLabel("Baud Rate:", Label.LEFT);
            baudLabel.setFont(font);
            portChoice = new JComboBox();
            portChoice.setFont(font);
            portChoice.addItem(new String("<Not Connected>"));
            portChoice.setSelectedIndex(0);
            portChoice.setEnabled(!portOpened);
            Enumeration ports = SerialService.getSerialPortIdentifiers();
            while (ports.hasMoreElements())
            {
               String strPort = (String)ports.nextElement();
               portChoice.addItem(strPort);
               if(strComPort!=null && strPort.toUpperCase().equals(strComPort.toUpperCase()))
                  portChoice.setSelectedItem(strPort);
            }

            baudChoice = new JComboBox();
            baudChoice.setFont(font);
            baudChoice.setActionCommand("baudrate");
            baudChoice.addItem("300");
            baudChoice.addItem("2400");
            baudChoice.addItem("9600");
            baudChoice.addItem("14400");
            baudChoice.addItem("19200");
            baudChoice.addItem("28800");
            baudChoice.addItem("38400");
            baudChoice.addItem("57600");
            baudChoice.addItem("115200");
            if(strBaud!=null)
               baudChoice.setSelectedItem(strBaud);
            else
               baudChoice.setSelectedItem(baud);
            baudChoice.addActionListener(this);
         configurationPanel.add(portNameLabel);
         configurationPanel.add(baudLabel);
         configurationPanel.add(portChoice);
         configurationPanel.add(baudChoice);
         JPanel buttonPanel = new JPanel();
            if(portOpened)
            {
               openCloseButton = new JButton("Close Port");
               openCloseButton.setActionCommand("closeport");
            }
            else
            {
               openCloseButton = new JButton("Open Port");
               openCloseButton.setActionCommand("openport");
            }
            openCloseButton.setFont(font);
            openCloseButton.addActionListener(this);

            resetButton = new JButton("Reset");
            resetButton.setFont(font);
            resetButton.setEnabled(portOpened);
            resetButton.addActionListener(this);
            resetButton.setActionCommand("Reset");

            JPanel dtrPanel = new JPanel();
            dtrPanel.setBorder(BorderFactory.createTitledBorder("DTR"));
               dtrSet = new JRadioButton("Set", false);
               dtrSet.setFont(font);
               dtrSet.setActionCommand("Set DTR");
               dtrSet.addActionListener(this);
               dtrSet.setSelected(true);
               dtrSet.setEnabled(portOpened);
               dtrClear = new JRadioButton("Clear", false);
               dtrClear.setFont(font);
               dtrClear.setActionCommand("Clear DTR");
               dtrClear.addActionListener(this);
               dtrClear.setEnabled(portOpened);
               ButtonGroup dtrGroup = new ButtonGroup();
               dtrGroup.add(dtrSet);
               dtrGroup.add(dtrClear);
            dtrPanel.add(dtrSet);
            dtrPanel.add(dtrClear);
         buttonPanel.add(openCloseButton);
         buttonPanel.add(resetButton);
         buttonPanel.add(dtrPanel);
      southPanel.add(buttonPanel, BorderLayout.SOUTH);
      southPanel.add(configurationPanel,BorderLayout.NORTH);
      content.add(southPanel,BorderLayout.SOUTH);

      JMenuBar menuBar = new JMenuBar();
      menuBar.setFont(font);
      this.setJMenuBar(menuBar);

      fileMenu = new JMenu("File");
      fileMenu.setMnemonic('F');
      fileMenu.setFont(font);
         exitItem = new JMenuItem("Exit");
         exitItem.setFont(font);
         exitItem.setActionCommand("exit");
         exitItem.addActionListener(this);
      fileMenu.add(exitItem);
      menuBar.add(fileMenu);

      editMenu = new JMenu("Edit");
      editMenu.setMnemonic('E');
      editMenu.setFont(font);
         JMenuItem selectAllItem = new JMenuItem("Select All");
         selectAllItem.setFont(font);
         selectAllItem.setActionCommand("select");
         selectAllItem.addActionListener(this);
         JMenuItem copyItem = new JMenuItem("Copy");
         copyItem.setFont(font);
         copyItem.setActionCommand("copy");
         copyItem.addActionListener(this);
         JMenuItem pasteItem = new JMenuItem("Paste");
         pasteItem.setFont(font);
         pasteItem.setActionCommand("paste");
         pasteItem.addActionListener(this);
         JMenuItem clearItem = new JMenuItem("Clear Window");
         clearItem.setFont(font);
         clearItem.setActionCommand("clear");
         clearItem.addActionListener(this);
      editMenu.add(clearItem);
      editMenu.add(selectAllItem);
      editMenu.add(copyItem);
      editMenu.add(pasteItem);
      menuBar.add(editMenu);

      loaderMenu = new JMenu("Loader");
      loaderMenu.setMnemonic('L');
      loaderMenu.setFont(font);
         loadHexFileItem = new JMenuItem("Load HEX");
         loadHexFileItem.setFont(font);
         loadHexFileItem.setActionCommand("load");
         loadHexFileItem.addActionListener(this);
         loadHexFileItem.setEnabled(portOpened);
         verifyHexFileItem = new JMenuItem("Verify HEX");
         verifyHexFileItem.setFont(font);
         verifyHexFileItem.setActionCommand("verify");
         verifyHexFileItem.addActionListener(this);
         verifyHexFileItem.setEnabled(portOpened);
      loaderMenu.add(loadHexFileItem);
      loaderMenu.add(verifyHexFileItem);
      menuBar.add(loaderMenu);

      optionsMenu = new JMenu("Options");
      optionsMenu.setMnemonic('O');
      optionsMenu.setFont(font);
         showRawMenuItem = new JMenuItem("Show Raw Messages");
         showRawMenuItem.setFont(font);
         showRawMenuItem.addActionListener(this);
         showRawMenuItem.setActionCommand("showraw");
         setMenu = new JMenu("Set Properties");
         setMenu.setFont(font);
            setServiceFilename = new JMenuItem("Set Account Filename");
            setServiceFilename.setFont(font);
            setServiceFilename.addActionListener(this);
            setServiceFilename.setActionCommand("set_F");
            setLCDText = new JMenuItem("Set LCD Text");
            setLCDText.setFont(font);
            setLCDText.addActionListener(this);
            setLCDText.setActionCommand("set_G");
         setMenu.add(setServiceFilename);
         setMenu.add(setLCDText);
         getMenu = new JMenu("Get Properties");
         getMenu.setFont(font);
      optionsMenu.add(showRawMenuItem);
      optionsMenu.add(setMenu);
      optionsMenu.add(getMenu);
      menuBar.add(optionsMenu);

      readText = new JTextArea("");
      readText.setFont(font);
      readText.addKeyListener(new terminalKeyListener());
      JScrollPane pane = new JScrollPane(readText);
      pane.setFont(font);
      content.add(pane, BorderLayout.CENTER);
   }


   public void unknownMessage(byte[] buffer, int offset, int count)
   {
      readText.append("UNKNOWN: " + new String(buffer, 0, count));
   }

   public void processMessage(byte[] buffer, int offset, int count)
      throws Convert.ConvertException
   {
      // check for error
      if(buffer[0]=='E' && buffer[1]=='E')
      {
         readText.append("Error: ");

         int errorCode = Convert.toInt(Convert.toByteArray(new String(buffer, 2, 2)), 0, 1);
         switch(errorCode)
         {
            case 0:
               readText.append("No Error Occurred");
               break;
            case 1:
               readText.append("No valid Token");
               break;
            case 2:
               readText.append("Tear occurred, Token not updated");
               break;
            case 3:
               readText.append("Tear occurred, Token state is unknown");
               break;
            case 4:
               readText.append("Tear occurred, Token needs verification");
               break;
            case 5:
               readText.append("Erase Scratchpad failed for Coprocessor device");
               break;
            case 6:
               readText.append("Failed to find money page on Token");
               break;
            case 7:
               readText.append("Could not get challenge from Coprocessor");
               break;
            case 8:
               readText.append("Could not issue challenge to Token for initial read");
               break;
            case 9:
               readText.append("Could not issue challenge to Token for verification");
               break;
            case 10:
               readText.append("Failed to read the account file of the Token");
               break;
            case 11:
               readText.append("Failed to compute the Unique Token Secret");
               break;
            case 12:
               readText.append("Token failed authentication");
               break;
            case 13:
               readText.append("Token purse file has invalid signature");
               break;
            case 14:
               readText.append("Failed to write the account page of Token");
               break;
            case 15:
               readText.append("Token has insufficient funds for transaction");
               break;
            case 16:
               readText.append("Failed to erase scratchpad of Token");
               break;
            default:
               unknownMessage(buffer, 0, count);
               break;
         }
      }
      else if(buffer[0]=='E')
      {
         readText.append("Event: ");
         int balance;
         switch(buffer[1])
         {
            case 'A':
               readText.append("Arrival: ");
               readText.append(Convert.toHexString(Convert.toByteArray(new String(buffer, 4, 16)), 0, 8, " "));
               break;
            case 'D':
               balance = Convert.toInt(Convert.toByteArray(new String(buffer, 4, 6)), 0, 3);
               readText.append("Debit performed, balance: $" + Convert.toString(balance/100d, 2));
               break;
            case 'L':
               readText.append("Departure: ");
               readText.append(Convert.toHexString(Convert.toByteArray(new String(buffer, 4, 16)), 0, 8, " "));
               break;
            case 'R':
               readText.append("Reset");
               break;
            case 'S':
               balance = Convert.toInt(Convert.toByteArray(new String(buffer, 4, 6)), 0, 3);
               readText.append("Token Authenticated, balance: $" + Convert.toString(balance/100d, 2));
               break;
            default:
               unknownMessage(buffer, 0, count);
               break;
         }
      }
      else
      {
         unknownMessage(buffer, 0, count);
      }

      readText.append("\r\n");
      int charCnt = readText.getDocument().getLength();
      if(charCnt>500000)
      {
         try
         {
            readText.setText(readText.getDocument().getText(100000, 400000));
         }
         catch(Exception ex){;}
      }
      readText.setCaretPosition(readText.getDocument().getEndPosition().getOffset() - 1);
      readText.repaint();
   }

   public void actionPerformed(ActionEvent ae)
   {
      String command = ae.getActionCommand();
      System.out.println("{" + command + "}");
      if(command.equals("exit"))
      {
         exitApplication();
      }
      else if(command.equals("load") || command.equals("verify"))
      {
         if(lastUsedDir==null)
            lastUsedDir = ".";
         if(command.equals("load"))
            chooser.setDialogTitle("Load Hex File...");
         else
            chooser.setDialogTitle("Verify Hex File...");
         chooser.setCurrentDirectory(new File(lastUsedDir));
         if(chooser.showOpenDialog(this)==JFileChooser.APPROVE_OPTION)
         {
            try
            {
               lastUsedDir = chooser.getCurrentDirectory().getCanonicalPath();
            }
            catch(IOException io)
            {;}
            File f = chooser.getSelectedFile();
            new Thread(new LoadHexFile(f, command.equals("load"))).start();
         }
      }
      else if(command.equals("Set DTR"))
      {
         serialService.setDTR(true);
      }
      else if(command.equals("Clear DTR"))
      {
         serialService.setDTR(false);
      }
      else if(command.equals("showraw"))
      {
         this.showRawMessages = true;
         showRawMenuItem.setText("Hide Raw Messages");
         showRawMenuItem.setActionCommand("hideraw");
      }
      else if(command.equals("hideraw"))
      {
         this.showRawMessages = false;
         showRawMenuItem.setText("Show Raw Messages");
         showRawMenuItem.setActionCommand("showraw");
      }
      else if(command.equals("baudrate"))
      {
         try
         {
            serialService.setBaudRate(
               Integer.parseInt(baudChoice.getSelectedItem().toString()));
         }
         catch(Exception e)
         {e.printStackTrace();}
      }
      else if(command.equals("openport"))
      {
         if(portChoice.getSelectedIndex()>0)
         {
            try
            {
               serialService = SerialService.getSerialService(
                  portChoice.getSelectedItem().toString());
               serialService.openPort(this);
               serialService.setBaudRate(
                  Integer.parseInt(baudChoice.getSelectedItem().toString()));

               loadHexFileItem.setEnabled(true);
               verifyHexFileItem.setEnabled(true);
               dtrSet.setEnabled(true);
               dtrClear.setEnabled(true);
               portChoice.setEnabled(false);
               openCloseButton.setText("Close Port");
               openCloseButton.setActionCommand("closeport");
            }
            catch(Exception e)
            {e.printStackTrace();}
         }
      }
      else if(command.equals("closeport"))
      {
         portOpened = false;

         serialService.closePort();

         loadHexFileItem.setEnabled(false);
         verifyHexFileItem.setEnabled(false);
         dtrSet.setEnabled(false);
         dtrClear.setEnabled(false);
         portChoice.setEnabled(true);
         openCloseButton.setText("Open Port");
         openCloseButton.setActionCommand("openport");
      }
      else if(command.indexOf("set_")==0)
      {
         switch(command.charAt(4))
         {
            // set the service filename for user token acct file
            case 'F':
               setServiceFilename();
               //memcpy(MoneyFileName, &SInBuf[2], 4);
               break;
            // Set the LCD text
            case 'G':
               setLCDText();
               break;
            // acknowledge a lock-step event
            case 'K':
               //setLockStepAck();
               break;
            // enable a lock-step event
            case 'L':
               //setLockStepEnable();
               break;
            // select a coprocessor for interaction
            case 'N':
               //CoprocessorNumber = SInBuf[2]=='1'?1:0;
               break;
            // Set the page read for coprocessor
            case 'P':
               //PageToReadCopr = hexToByte(&SInBuf[2]);
               break;
            // soft reset the ecash processor
            case 'R':
               //softReset();
               break;
            // Set the page read for user
            case 'U':
               //PageToReadUser = hexToByte(&SInBuf[2]);
               break;
            // play the beeper
            case 'S':
               //beep(hexToByte(&SInBuf[2]));*/
               break;
         }
      }
   }

   private void setServiceFilename()
   {
      JPanel inputPanel = new JPanel();
      inputPanel.setLayout(new BoxLayout(inputPanel, BoxLayout.X_AXIS));
      inputPanel.add(new JLabel("Account Filename: "));
      JTextField tf = new JTextField("DLSM.102", 10);
      inputPanel.add(tf);
      if(JOptionPane.showConfirmDialog(this, inputPanel, "Set Account Filename",
            JOptionPane.OK_CANCEL_OPTION) == JOptionPane.OK_OPTION)
      {
         String filename = tf.getText();
         int index = filename.indexOf('.');
         if((index==-1 && filename.length()==4) || index==4)
         {
            filename = filename.substring(0,4);
            try
            {
               serialService.setNotifyOnDataAvailable(false);
               serialService.write("SF" + filename + "\r\n");
               while(serialService.available()==0)
                  Thread.yield();
               serialService.flush();
               serialService.setNotifyOnDataAvailable(true);
            }
            catch(IOException ioe)
            {
               ioe.printStackTrace();
            }
         }
      }
   }

   private void setLCDText()
   {
      JPanel inputPanel = new JPanel();
      inputPanel.setLayout(new BoxLayout(inputPanel, BoxLayout.X_AXIS));
      inputPanel.add(new JLabel("LCD Text: "));
      JTextField tf = new JTextField("eCash ver.8", 16);
      inputPanel.add(tf);
      if(JOptionPane.showConfirmDialog(this, inputPanel, "Set LCD Text",
            JOptionPane.OK_CANCEL_OPTION) == JOptionPane.OK_OPTION)
      {
         try
         {
            String text = tf.getText();
            int length = text.length();
            if(length>16)
            {
               length = 16;
               text = text.substring(0, 16);
            }
            serialService.setNotifyOnDataAvailable(false);

            System.out.print("SG" + byteToHex((byte)length) + text + "\r\n");
            serialService.write("SG" + byteToHex((byte)length) + text + "\r\n");
            while(serialService.available()==0)
               Thread.yield();
            serialService.flush();
            serialService.setNotifyOnDataAvailable(true);
         }
         catch(IOException ioe)
         {
            ioe.printStackTrace();
         }
      }
   }

   private String byteToHex(byte b)
   {
      char[] str = new char[2];
      str[1] = (char)(((b&0x0F)>9)?(b&0x0F)-10+'A':(b&0x0F)+'0');
      b = (byte)((b&0xF0)>>4);
      str[0] = (char)((b>9)?b-10+'A':b+'0');
      return new String(str);
   }

   private final byte[] read_buffer = new byte[128];
   private int read_cnt = 0;
   public void serialEvent(SerialPortEvent spe)
   {
      if(spe.getEventType()!=SerialPortEvent.DATA_AVAILABLE)
         return;

      try
      {
         int available;
         while((available=serialService.available())>0)
         {
            int readLength = Math.min(available,
                                      read_buffer.length-read_cnt);
            read_cnt += serialService.read(read_buffer, read_cnt, readLength);
            if(read_cnt>0)
            {
               if(showRawMessages)
               {
                  readText.append(new String(read_buffer, 0, read_cnt));
                  read_cnt = 0;
                  readText.setCaretPosition(readText.getDocument().getEndPosition().getOffset() - 1);
                  readText.repaint();
               }
               else
               {
                  int end;
                  do
                  {
                     end = -1;
                     // look for \n, the line terminator
                     for(int i=0; i<read_cnt; i++)
                     {
                        if(read_buffer[i]=='\n')
                        {
                           end = i;
                           break;
                        }
                     }
                     if(end!=-1)
                     {
                        try
                        {
                           processMessage(read_buffer, 0, end);
                        }
                        catch(Convert.ConvertException ce)
                        {
                           ce.printStackTrace();
                        }

                        if(read_cnt>end+1)
                        {
                           System.arraycopy(read_buffer, end+1, read_buffer, 0,
                                            read_cnt-(end+1));
                           read_cnt -= (end+1);
                        }
                        else
                        {
                           read_cnt = 0;
                        }
                     }
                  }
                  while(end!=-1);
               }
               String text = readText.getText();
               if(text.length()> 10000)
               {
                  readText.setText(text.substring(6000));
                  readText.setCaretPosition(readText.getDocument().getEndPosition().getOffset() - 1);
                  readText.repaint();
               }

            }
         }
      }
      catch(Exception e) {e.printStackTrace();}
   }

   private class HexFileFilter extends javax.swing.filechooser.FileFilter
   {
      public boolean accept(File f)
      {
         return f.isDirectory() || f.getName().toUpperCase().endsWith(".HEX");
      }
      public String getDescription()
      {
         return "HEX Files (*.hex)";
      }
   }

   private class LoadHexFile implements Runnable
   {
      File file = null;
      boolean load = true;

      public LoadHexFile(File file, boolean load)
      {
         this.file = file;
         this.load = load;
      }

      public void run()
      {
         final byte[] read_buffer = new byte[128];
         int cnt = 0;
         serialService.setNotifyOnDataAvailable(false);
         try
         {
            dtrSet.setSelected(true);
            if(serialService.isDTR())
               serialService.setDTR(false);
            serialService.setDTR(true);
            serialService.write("\r\n\r\n");
            serialService.flush();
            cnt = 0;

            if(load)
            {
               // Klear the memory
               serialService.write("K\r\n");
               boolean waitOnClear = true;
               while(waitOnClear)
               {
                  cnt += serialService.read(read_buffer, cnt, read_buffer.length-cnt);
                  if(cnt>0)
                  {
                     String strRead = new String(read_buffer, 0, cnt);
                     if(strRead.indexOf("K\r\n\r\n> ")>=0)
                        waitOnClear = false;
                  }
                  WAITMS(100);
               }
            }

            if(load)
               serialService.write("L\r\n");
            else
               serialService.write("V\r\n");
            boolean waitOnLoad = true;
            while(waitOnLoad)
            {
               cnt += serialService.read(read_buffer, cnt, read_buffer.length-cnt);
               if(cnt>0)
               {
                  String strRead = new String(read_buffer, 0, cnt);
                  if(strRead.indexOf("L\r\n")>=0 || strRead.indexOf("V\r\n")>=0)
                     waitOnLoad = false;
               }
               WAITMS(100);
            }

            serialService.flush();
            if(load)
               readText.setText("Loading ");
            else
               readText.setText("Verifying ");
            readText.append("File: " + file.getAbsolutePath() + "\r\nDated: " + new Date(file.lastModified())+"\r\n");
            readText.repaint();
            java.io.FileReader fr = new java.io.FileReader(file);
            java.io.BufferedReader br = new java.io.BufferedReader(fr);
            String line;
            cnt = 0;
            while((line=br.readLine())!=null)
            {
               serialService.write(line);
               serialService.write("\r\n");
               char read;
               do
               {
                  read = (char)serialService.read();
               }
               while(!Character.isLetter(read)); //|| read==-1 || read=='>' || read==' ' || read=='\r' || read=='\n');
               readText.append("" + (char)read);
               if(++cnt==64)
               {
                  cnt = 0;
                  readText.append("\r\n");
               }
               readText.repaint();
               if(read!='G')
               {
                  readText.append("\r\nError\r\n{" + read + "}" + load);
                  break;
               }
            }
            readText.append("\r\nDone Loading\r\n");
            serialService.flush();
            cnt = 0;
         }
         catch(Exception e) {e.printStackTrace();}

         serialService.setNotifyOnDataAvailable(true);
      }
   }

   private class terminalKeyListener extends KeyAdapter
   {
      public void keyPressed(KeyEvent e)
      {
         if (e.getModifiers() == ActionEvent.ALT_MASK)
         {
             menuBar.grabFocus();
         }

         int keyCode = e.getKeyCode();
         try
         {
            if(keyCode==KeyEvent.VK_ENTER)
            {
               //synchronized(serialService)
               {
                  serialService.write("\r\n");
                  if(!serialService.isDTR())
                  {
                     readText.append("\r\n");
                  }
               }
            }
            else if ((keyCode != KeyEvent.VK_SHIFT) && (keyCode != KeyEvent.VK_ALT) &&
                     (keyCode != KeyEvent.VK_CONTROL) && (keyCode != KeyEvent.VK_CAPS_LOCK))
            {
               //synchronized(serialService)
               {
                  serialService.write(e.getKeyChar());
                  if(!serialService.isDTR())
                  {
                     readText.append(""+e.getKeyChar());
                  }
               }
            }
         }
         catch(Exception ex){ex.printStackTrace();}
         e.consume();
      }

      public void keyReleased(KeyEvent e)
      {
         e.consume();
         return;
      }
      public void keyTyped(KeyEvent e)
      {
         e.consume();
         return;
      }
   }

   /**
   * Load the initialization file and set the preferences.
   */
   private void loadINIFile()
   {
      //Get the last used directory from our temp file if it exists...
      try
      {
         FileInputStream file = new FileInputStream(MACOSX ? ".javakit" : "JavaKit.tmp");
         BufferedReader data = new BufferedReader(new InputStreamReader((InputStream)file));
         lastUsedDir = new String(data.readLine());
         baud = new String(data.readLine());
         width = Integer.parseInt(data.readLine(), 16);
         height = Integer.parseInt(data.readLine(), 16);
         xPos = Integer.parseInt(data.readLine(), 16);
         yPos = Integer.parseInt(data.readLine(), 16);
      }
      catch(Exception e)
      {
         lastUsedDir = System.getProperty("user.dir");
      }
   }

   /**
   * Write the initialization file.
   */
   private void writeINIFile()
   {
      //Write the new current directory to our temp file...
      try
      {
         FileOutputStream outFile = new FileOutputStream(MACOSX ? ".javakit" : "JavaKit.tmp");
         DataOutputStream dataOut = new DataOutputStream((OutputStream)outFile);
         dataOut.writeBytes(lastUsedDir + "\n");
         dataOut.writeBytes(baudChoice.getSelectedItem() + "\n");
         Dimension dim = getSize();
         dataOut.writeBytes(Integer.toHexString(dim.width) + "\n");
         dataOut.writeBytes(Integer.toHexString(dim.height) + "\n");
         Point point = getLocation();
         dataOut.writeBytes(Integer.toHexString(point.x) + "\n");
         dataOut.writeBytes(Integer.toHexString(point.y) + "\n");
      }
      catch(IOException ioe)
      {
         //DRAIN
      }
   }

   private void exitApplication()
   {
      System.out.println("Closing");
      if(serialService!=null && serialService.isPortOpen())
      {
         serialService.closePort();
      }
      writeINIFile();
      System.exit(0);
   }

   /**
    * prints a message and exits the application.
    */
   private void ABORT(String msg)
   {
      //System.out.println("Aborting program: " + msg);
      serialService.closePort();
      writeINIFile();
      System.exit(1);
   }

   /**
    * An easy thread.sleep method.
    */
   static final void WAITMS(final long ms)
   {
      try
      {
         Thread.sleep(ms);
         Thread.yield();
      }
      catch(InterruptedException ie)
      {;}
   }

   /**
    * Method printUsageString
    *
    */
   public static void printUsageString ()
   {
      System.out.println(
         "eCash Firmware Monitor\r\n");
      System.out.println("Usage: ");
      System.out.println("   java eCashMonitor \r\n");
      System.exit(1);
   }

   public static void main(String[] args)
   {
        String comPort = null;
        String baud = null;
        boolean log = false;
        boolean debugMode = false;

        for (int i=0; i<args.length; i++)
        {
            if (args[i].equalsIgnoreCase("-port"))
            {
                comPort = args[++i];
            }
            else if (args[i].equalsIgnoreCase("-baud"))
            {
                baud = args[++i];
            }
            else
            {
                //System.out.println("\nUnknown option: " + args[i]);
                printUsageString();
                System.exit(1);
            }
        }

        if (!System.getProperty("os.name").startsWith("Mac OS X"))
        {
          //Attempt to set the swing (JFC) "Look And Feel".
          boolean lookAndFeelSet = false;
          int attempt = 0;
          //Loop through all of the possible look and feels until we find one that works.
          while(!lookAndFeelSet)
          {
            try
            {
                switch (attempt++)
                {
                    case 0:
                        //Try the Metal "Look And Feel" first.
                        UIManager.setLookAndFeel("com.sun.java.swing.plaf.metal.MetalLookAndFeel");
                    break;

                    case 1:
                        //Try to load the Windows "Look And Feel" next.
                        UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
                    break;

                    case 2:
                        //Try the Motif "Look And Feel" next.
                        UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");
                    break;

                    case 3:
                        //Finally, try the Basic "Look And Feel".
                        UIManager.setLookAndFeel("com.sun.java.swing.plaf.basic.BasicLookAndFeel");
                    break;
                }
                lookAndFeelSet = true;
            }
            catch(Exception e)
            {;}
          }
        }

      //JavaKit javaKit = new JavaKit(comPort, macroFiles, baud, padSize, binPause,
      //                              exitAfterRun, bankSize, romSize, log, debugMode);
      //javaKit.setVisible(true);
      //javaKit.repaint();

      try
      {
         eCashMonitor mon = new eCashMonitor(comPort, baud);
         mon.setVisible(true);
      }
      catch(Exception e)
      {e.printStackTrace();}
   }
}
